# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import logging
import os
import subprocess
from textwrap import dedent
from unittest import mock

import fixtures
import pytest
from testtools.matchers import Contains, Equals, FileContains, HasLength

import snapcraft
from snapcraft import storeapi
from snapcraft.internal import errors, meta
from snapcraft.plugins.v1 import kernel

from . import PluginsV1BaseTestCase


class KernelPluginTestCase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class Options:
            build_parameters = []
            kconfigfile = None
            kconfigflavour = None
            kdefconfig = []
            kconfigs = []
            kernel_image_target = "bzImage"
            kernel_with_firmware = True
            kernel_initrd_modules = []
            kernel_initrd_firmware = []
            kernel_device_trees = []
            kernel_initrd_compression = "gz"
            build_attributes = []

        self.options = Options()

        patcher = mock.patch("subprocess.check_call")
        self.check_call_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch.object(kernel.KernelPlugin, "run")
        self.run_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch.object(kernel.KernelPlugin, "run_output")
        self.run_output_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("snapcraft.BasePlugin.build")
        self.base_build_mock = patcher.start()
        self.addCleanup(patcher.stop)

        @contextlib.contextmanager
        def tempdir():
            self.tempdir = "temporary-directory"
            os.mkdir(self.tempdir)
            yield self.tempdir

        patcher = mock.patch("tempfile.TemporaryDirectory")
        self.tempdir_mock = patcher.start()
        self.tempdir_mock.side_effect = tempdir
        self.addCleanup(patcher.stop)

    def test_schema(self):
        schema = kernel.KernelPlugin.schema()

        properties = schema["properties"]
        self.assertThat(properties["kdefconfig"]["type"], Equals("array"))
        self.assertThat(properties["kdefconfig"]["default"], Equals(["defconfig"]))

        self.assertThat(properties["kconfigfile"]["type"], Equals("string"))
        self.assertThat(properties["kconfigfile"]["default"], Equals(None))

        self.assertThat(properties["kconfigflavour"]["type"], Equals("string"))
        self.assertThat(properties["kconfigflavour"]["default"], Equals(None))

        for prop in [
            "kconfigs",
            "kernel-initrd-modules",
            "kernel-initrd-firmware",
            "kernel-device-trees",
        ]:
            self.assertThat(properties[prop]["type"], Equals("array"))
            self.assertThat(properties[prop]["default"], Equals([]))
            self.assertThat(properties[prop]["minitems"], Equals(1))
            self.assertThat(properties[prop]["items"]["type"], Equals("string"))
            self.assertTrue(properties[prop]["uniqueItems"])

        self.assertThat(
            properties["kernel-image-target"]["oneOf"],
            Equals([{"type": "string"}, {"type": "object"}]),
        )
        self.assertThat(properties["kernel-image-target"]["default"], Equals(""))

        self.assertThat(properties["kernel-with-firmware"]["type"], Equals("boolean"))
        self.assertThat(properties["kernel-with-firmware"]["default"], Equals(True))

        self.assertThat(
            properties["kernel-initrd-compression"]["type"], Equals("string")
        )
        self.assertThat(
            properties["kernel-initrd-compression"]["default"], Equals("gz")
        )
        self.assertThat(properties["kernel-initrd-compression"]["enum"], Equals(["gz"]))

    def test_get_build_properties(self):
        expected_build_properties = [
            "kernel-image-target",
            "kernel-with-firmware",
            "kernel-initrd-modules",
            "kernel-initrd-firmware",
            "kernel-device-trees",
            "kernel-initrd-compression",
        ]
        resulting_build_properties = kernel.KernelPlugin.get_build_properties()
        expected_build_properties.extend(
            snapcraft.plugins.v1.kbuild.KBuildPlugin.get_build_properties()
        )

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)

    def _assert_generic_check_call(self, builddir, installdir, os_snap_path):
        self.assertThat(self.check_call_mock.call_count, Equals(4))
        self.check_call_mock.assert_has_calls(
            [
                mock.call('yes "" | make -j2 oldconfig', shell=True, cwd=builddir),
                mock.call(
                    ["unsquashfs", os_snap_path, "boot"], cwd="temporary-directory"
                ),
                mock.call(
                    "cat temporary-directory/squashfs-root/boot"
                    "/initrd.img-core | xz -dc | cpio -i",
                    cwd=os.path.join(builddir, "initrd-staging"),
                    shell=True,
                ),
                mock.call(
                    "find . | cpio --create --format=newc | "
                    "gzip > {}".format(os.path.join(installdir, "initrd-4.4.2.img")),
                    cwd=os.path.join(builddir, "initrd-staging"),
                    shell=True,
                ),
            ]
        )

    def _assert_common_assets(self, installdir):
        for asset in [
            "initrd-4.4.2.img",
            "initrd.img",
            "kernel.img",
            "bzImage-4.4.2",
            "System.map-4.4.2",
        ]:
            self.assertTrue(
                os.path.exists(os.path.join(installdir, asset)),
                "Missing {}".format(asset),
            )

    def _simulate_build(
        self,
        sourcedir,
        builddir,
        installdir,
        do_dtbs=False,
        do_release=True,
        do_kernel=True,
        do_system_map=True,
        do_firmware=True,
    ):
        os.makedirs(sourcedir)
        kernel_version = "4.4.2"

        def create_assets():
            build_arch_path = os.path.join(
                builddir, "arch", self.project.kernel_arch, "boot"
            )

            initrd_path = os.path.join(
                installdir, "initrd-{}.img".format(kernel_version)
            )
            release_path = os.path.join(builddir, "include", "config", "kernel.release")
            kernel_path = os.path.join(
                build_arch_path, self.options.kernel_image_target
            )
            system_map_path = os.path.join(builddir, "System.map")
            modules_dep_path = os.path.join(
                installdir, "lib", "modules", kernel_version, "modules.dep"
            )
            modules_dep_bin_path = "{}.bin".format(modules_dep_path)
            if do_firmware:
                fw_bin_path = os.path.join(installdir, "lib", "firmware", "fake-fw.bin")
                os.makedirs(os.path.join(installdir, "lib", "firmware", "fake-fw-dir"))
            dtb_path = os.path.join(build_arch_path, "dts", "fake-dtb.dtb")

            os.makedirs(os.path.dirname(release_path))
            with open(release_path, "w") as f:
                if do_release:
                    f.write("{}\n".format(kernel_version))
                else:
                    f.write("\n")

            files = [initrd_path, modules_dep_path, modules_dep_bin_path]
            if do_kernel:
                files.append(kernel_path)
            if do_system_map:
                files.append(system_map_path)
            if do_dtbs:
                files.append(dtb_path)
            if do_firmware:
                files.append(fw_bin_path)

            for f in files:
                os.makedirs(os.path.dirname(f), exist_ok=True)
                open(f, "w").close()

        self.base_build_mock.side_effect = create_assets

    def test_unpack_gzip_initrd(self):
        def check_call_effect(*args, **kwargs):
            if "xz" in args[0]:
                raise subprocess.CalledProcessError(1, args)

        self.check_call_mock.side_effect = check_call_effect

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)
        plugin._unpack_generic_initrd()

        self.check_call_mock.assert_has_calls(
            [
                mock.call(
                    "cat temporary-directory/squashfs-root/boot/"
                    "initrd.img-core | gzip -dc | cpio -i",
                    cwd=os.path.join(plugin.builddir, "initrd-staging"),
                    shell=True,
                )
            ]
        )

    def test_unpack_lzma_initrd(self):
        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        plugin._unpack_generic_initrd()

        self.check_call_mock.assert_has_calls(
            [
                mock.call(
                    "cat temporary-directory/squashfs-root/boot/"
                    "initrd.img-core | xz -dc | cpio -i",
                    cwd=os.path.join(plugin.builddir, "initrd-staging"),
                    shell=True,
                )
            ]
        )

    def test_unpack_unsupported_initrd_type(self):
        def check_call_effect(*args, **kwargs):
            if "unsquashfs" not in args[0]:
                raise subprocess.CalledProcessError(1, args)

        self.check_call_mock.side_effect = check_call_effect
        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self.assertRaises(RuntimeError, plugin._unpack_generic_initrd)

    def test_pack_initrd_modules(self):
        self.options.kernel_initrd_modules = ["squashfs", "vfat"]

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        # Fake some assets
        plugin.kernel_release = "4.4"
        modules_path = os.path.join(plugin.installdir, "lib", "modules", "4.4")
        initrd_modules_staging_path = os.path.join("staging", "lib", "modules", "4.4")
        os.makedirs(modules_path)
        open(os.path.join(modules_path, "modules.dep"), "w").close()
        open(os.path.join(modules_path, "modules.dep.bin"), "w").close()
        os.makedirs(initrd_modules_staging_path)
        open(os.path.join(plugin.installdir, "initrd-4.4.img"), "w").close()

        with mock.patch.object(plugin, "_unpack_generic_initrd") as m_unpack:
            m_unpack.return_value = "staging"
            plugin._make_initrd()

        modprobe_cmd = [
            "modprobe",
            "-n",
            "--show-depends",
            "-d",
            plugin.installdir,
            "-S",
            "4.4",
        ]
        self.run_output_mock.assert_has_calls(
            [mock.call(modprobe_cmd + ["squashfs"], env=mock.ANY)]
        )
        self.run_output_mock.assert_has_calls(
            [mock.call(modprobe_cmd + ["vfat"], env=mock.ANY)]
        )

    def test_pack_initrd_modules_return_same_deps(self):
        self.options.kernel_initrd_modules = ["squashfs", "vfat"]

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        # Fake some assets
        plugin.kernel_release = "4.4"
        modules_path = os.path.join(plugin.installdir, "lib", "modules", "4.4")
        initrd_modules_staging_path = os.path.join("staging", "lib", "modules", "4.4")
        os.makedirs(modules_path)
        open(os.path.join(modules_path, "modules.dep"), "w").close()
        open(os.path.join(modules_path, "modules.dep.bin"), "w").close()
        os.makedirs(initrd_modules_staging_path)
        open(os.path.join(plugin.installdir, "initrd-4.4.img"), "w").close()
        open(os.path.join(plugin.installdir, "serport.ko"), "w").close()

        self.run_output_mock.return_value = "insmod {}/serport.ko".format(
            plugin.installdir
        )

        with mock.patch.object(plugin, "_unpack_generic_initrd") as m_unpack:
            m_unpack.return_value = "staging"
            plugin._make_initrd()

        modprobe_cmd = [
            "modprobe",
            "-n",
            "--show-depends",
            "-d",
            plugin.installdir,
            "-S",
            "4.4",
        ]
        self.run_output_mock.assert_has_calls(
            [mock.call(modprobe_cmd + ["squashfs"], env=mock.ANY)]
        )
        self.run_output_mock.assert_has_calls(
            [mock.call(modprobe_cmd + ["vfat"], env=mock.ANY)]
        )

    @mock.patch.object(snapcraft.ProjectOptions, "kernel_arch", new="not_arm")
    def test_build_with_kconfigfile(self):
        self.options.kconfigfile = "config"
        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self._simulate_build(plugin.sourcedir, plugin.builddir, plugin.installdir)

        plugin.build()

        self._assert_generic_check_call(
            plugin.builddir, plugin.installdir, plugin.os_snap
        )

        self.assertThat(self.run_mock.call_count, Equals(2))
        self.run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2", "bzImage", "modules"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "modules_install",
                        "INSTALL_MOD_PATH={}".format(plugin.installdir),
                        "firmware_install",
                        "INSTALL_FW_PATH={}".format(
                            os.path.join(plugin.installdir, "lib", "firmware")
                        ),
                    ]
                ),
            ]
        )

        config_file = os.path.join(plugin.builddir, ".config")
        self.assertTrue(os.path.exists(config_file))

        with open(config_file) as f:
            config_contents = f.read()

        self.assertThat(config_contents, Equals("ACCEPT=y\n"))
        self._assert_common_assets(plugin.installdir)

    @mock.patch.object(snapcraft.ProjectOptions, "kernel_arch", new="not_arm")
    def test_build_verbose_with_kconfigfile(self):
        fake_logger = fixtures.FakeLogger(level=logging.DEBUG)
        self.useFixture(fake_logger)

        self.options.kconfigfile = "config"
        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self._simulate_build(plugin.sourcedir, plugin.builddir, plugin.installdir)

        plugin.build()

        self.assertThat(self.check_call_mock.call_count, Equals(4))
        self.check_call_mock.assert_has_calls(
            [
                mock.call(
                    'yes "" | make -j2 V=1 oldconfig', shell=True, cwd=plugin.builddir
                ),
                mock.call(
                    ["unsquashfs", plugin.os_snap, "boot"], cwd="temporary-directory"
                ),
                mock.call(
                    "cat temporary-directory/squashfs-root/boot/"
                    "initrd.img-core | xz -dc | cpio -i",
                    cwd=os.path.join(plugin.builddir, "initrd-staging"),
                    shell=True,
                ),
                mock.call(
                    "find . | cpio --create --format=newc | "
                    "gzip > {}".format(
                        os.path.join(plugin.installdir, "initrd-4.4.2.img")
                    ),
                    cwd=os.path.join(plugin.builddir, "initrd-staging"),
                    shell=True,
                ),
            ]
        )

        self.assertThat(self.run_mock.call_count, Equals(2))
        self.run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2", "V=1", "bzImage", "modules"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "V=1",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "modules_install",
                        "INSTALL_MOD_PATH={}".format(plugin.installdir),
                        "firmware_install",
                        "INSTALL_FW_PATH={}".format(
                            os.path.join(plugin.installdir, "lib", "firmware")
                        ),
                    ]
                ),
            ]
        )

        config_file = os.path.join(plugin.builddir, ".config")
        self.assertTrue(os.path.exists(config_file))

        with open(config_file) as f:
            config_contents = f.read()

        self.assertThat(config_contents, Equals("ACCEPT=y\n"))
        self._assert_common_assets(plugin.installdir)

    @mock.patch.object(snapcraft.ProjectOptions, "kernel_arch", new="not_arm")
    def test_check_config(self):
        fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(fake_logger)

        self.options.kconfigfile = "config"
        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self._simulate_build(plugin.sourcedir, plugin.builddir, plugin.installdir)

        builtin, modules = plugin._do_parse_config(self.options.kconfigfile)
        plugin._do_check_config(builtin, modules)

        required_opts = (
            kernel.required_generic
            + kernel.required_security
            + kernel.required_snappy
            + kernel.required_systemd
        )
        for warn in required_opts:
            self.assertIn("CONFIG_{}".format(warn), fake_logger.output)

    @mock.patch.object(snapcraft.ProjectOptions, "kernel_arch", new="not_arm")
    def test_check_initrd(self):
        fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(fake_logger)

        self.options.kconfigfile = "config"
        self.options.kernel_initrd_modules = ["my-fake-module"]
        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self._simulate_build(plugin.sourcedir, plugin.builddir, plugin.installdir)

        builtin, modules = plugin._do_parse_config(self.options.kconfigfile)
        plugin._do_check_initrd(builtin, modules)

        for module in kernel.required_boot:
            self.assertIn("CONFIG_{}".format(module.upper()), fake_logger.output)

    @mock.patch.object(snapcraft.ProjectOptions, "kernel_arch", new="not_arm")
    def test_build_with_kconfigfile_and_kconfigs(self):
        self.options.kconfigfile = "config"
        self.options.kconfigs = ["SOMETHING=y", "ACCEPT=n"]

        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self._simulate_build(plugin.sourcedir, plugin.builddir, plugin.installdir)

        plugin.build()

        self._assert_generic_check_call(
            plugin.builddir, plugin.installdir, plugin.os_snap
        )

        self.assertThat(self.run_mock.call_count, Equals(2))
        self.run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2", "bzImage", "modules"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "modules_install",
                        "INSTALL_MOD_PATH={}".format(plugin.installdir),
                        "firmware_install",
                        "INSTALL_FW_PATH={}".format(
                            os.path.join(plugin.installdir, "lib", "firmware")
                        ),
                    ]
                ),
            ]
        )

        config_file = os.path.join(plugin.builddir, ".config")
        self.assertTrue(os.path.exists(config_file))

        with open(config_file) as f:
            config_contents = f.read()

        expected_config = """SOMETHING=y
ACCEPT=n

ACCEPT=y

SOMETHING=y
ACCEPT=n
"""
        self.assertThat(config_contents, Equals(expected_config))
        self._assert_common_assets(plugin.installdir)

    @mock.patch.object(snapcraft.ProjectOptions, "kernel_arch", new="not_arm")
    def test_build_with_defconfig_and_kconfigs(self):
        self.options.kdefconfig = ["defconfig"]
        self.options.kconfigs = ["SOMETHING=y", "ACCEPT=n"]

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        config_file = os.path.join(plugin.builddir, ".config")

        def fake_defconfig(*args, **kwargs):
            if os.path.exists(config_file):
                return
            with open(config_file, "w") as f:
                f.write("ACCEPT=y\n")

        self.run_mock.side_effect = fake_defconfig

        self._simulate_build(plugin.sourcedir, plugin.builddir, plugin.installdir)

        plugin.build()

        self._assert_generic_check_call(
            plugin.builddir, plugin.installdir, plugin.os_snap
        )

        self.assertThat(self.run_mock.call_count, Equals(3))
        self.run_mock.assert_has_calls(
            [
                mock.call(["make", "-j1", "defconfig"]),
                mock.call(["make", "-j2", "bzImage", "modules"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "modules_install",
                        "INSTALL_MOD_PATH={}".format(plugin.installdir),
                        "firmware_install",
                        "INSTALL_FW_PATH={}".format(
                            os.path.join(plugin.installdir, "lib", "firmware")
                        ),
                    ]
                ),
            ]
        )

        self.assertTrue(os.path.exists(config_file))

        with open(config_file) as f:
            config_contents = f.read()

        expected_config = """SOMETHING=y
ACCEPT=n

ACCEPT=y

SOMETHING=y
ACCEPT=n
"""
        self.assertThat(config_contents, Equals(expected_config))
        self._assert_common_assets(plugin.installdir)

    @mock.patch.object(snapcraft.ProjectOptions, "kernel_arch", new="not_arm")
    def test_build_with_two_defconfigs(self):
        self.options.kdefconfig = ["defconfig", "defconfig2"]

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        config_file = os.path.join(plugin.builddir, ".config")

        def fake_defconfig(*args, **kwargs):
            if os.path.exists(config_file):
                return
            with open(config_file, "w") as f:
                f.write("ACCEPT=y\n")

        self.run_mock.side_effect = fake_defconfig

        self._simulate_build(plugin.sourcedir, plugin.builddir, plugin.installdir)

        plugin.build()

        self._assert_generic_check_call(
            plugin.builddir, plugin.installdir, plugin.os_snap
        )

        self.assertThat(self.run_mock.call_count, Equals(3))
        self.run_mock.assert_has_calls(
            [
                mock.call(["make", "-j1", "defconfig", "defconfig2"]),
                mock.call(["make", "-j2", "bzImage", "modules"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "modules_install",
                        "INSTALL_MOD_PATH={}".format(plugin.installdir),
                        "firmware_install",
                        "INSTALL_FW_PATH={}".format(
                            os.path.join(plugin.installdir, "lib", "firmware")
                        ),
                    ]
                ),
            ]
        )

        self.assertTrue(os.path.exists(config_file))
        self._assert_common_assets(plugin.installdir)

    @mock.patch.object(snapcraft.ProjectOptions, "kernel_arch", new="not_arm")
    def test_build_with_kconfigfile_and_dtbs(self):
        self.options.kconfigfile = "config"
        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")
        self.options.kernel_device_trees = ["fake-dtb"]

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self._simulate_build(
            plugin.sourcedir, plugin.builddir, plugin.installdir, do_dtbs=True
        )

        plugin.build()

        self._assert_generic_check_call(
            plugin.builddir, plugin.installdir, plugin.os_snap
        )

        self.assertThat(self.run_mock.call_count, Equals(2))
        self.run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2", "bzImage", "modules", "fake-dtb.dtb"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "modules_install",
                        "INSTALL_MOD_PATH={}".format(plugin.installdir),
                        "firmware_install",
                        "INSTALL_FW_PATH={}".format(
                            os.path.join(plugin.installdir, "lib", "firmware")
                        ),
                    ]
                ),
            ]
        )

        config_file = os.path.join(plugin.builddir, ".config")
        self.assertTrue(os.path.exists(config_file))

        with open(config_file) as f:
            config_contents = f.read()

        self.assertThat(config_contents, Equals("ACCEPT=y\n"))
        self._assert_common_assets(plugin.installdir)
        self.assertTrue(
            os.path.exists(os.path.join(plugin.installdir, "dtbs", "fake-dtb.dtb"))
        )

    def test_build_with_kconfigfile_and_dtbs_not_found(self):
        self.options.kconfigfile = "config"
        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")
        self.options.kernel_device_trees = ["fake-dtb"]

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self._simulate_build(plugin.sourcedir, plugin.builddir, plugin.installdir)

        raised = self.assertRaises(RuntimeError, plugin.build)

        self.assertThat(
            str(raised), Equals("No match for dtb 'fake-dtb.dtb' was found")
        )

    @mock.patch.object(snapcraft.ProjectOptions, "kernel_arch", new="not_arm")
    def test_build_with_kconfigfile_and_modules(self):
        self.options.kconfigfile = "config"
        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")
        self.options.kernel_initrd_modules = ["my-fake-module"]

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self._simulate_build(plugin.sourcedir, plugin.builddir, plugin.installdir)

        def fake_unpack(*args, **kwargs):
            modules_dir = os.path.join(
                plugin.builddir, "initrd-staging", "lib", "modules", "4.4.2"
            )
            if os.path.exists(modules_dir):
                return
            os.makedirs(modules_dir)

        self.check_call_mock.side_effect = fake_unpack

        def fake_output(*args, **kwargs):
            if args[0][:3] == ["modprobe", "-n", "--show-depends"]:
                module_path = os.path.join(
                    plugin.installdir, "lib", "modules", "4.4.2", "some-module.ko"
                )
                open(module_path, "w").close()
                return "insmod {} enable_fbdev=1\n".format(module_path)
            else:
                raise Exception(args[0])

        self.run_output_mock.side_effect = fake_output

        # Set a path that doesn't contain '/sbin' so we can verify that it's
        # added to the modprobe call.
        self.useFixture(fixtures.EnvironmentVariable("PATH", "/usr/bin"))

        plugin.build()

        self._assert_generic_check_call(
            plugin.builddir, plugin.installdir, plugin.os_snap
        )

        self.assertThat(self.run_mock.call_count, Equals(2))
        self.run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2", "bzImage", "modules"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "modules_install",
                        "INSTALL_MOD_PATH={}".format(plugin.installdir),
                        "firmware_install",
                        "INSTALL_FW_PATH={}".format(
                            os.path.join(plugin.installdir, "lib", "firmware")
                        ),
                    ]
                ),
            ]
        )

        class _check_env:
            def __init__(self, test):
                self.test = test

            def __eq__(self, other):
                self.test.assertThat(other, Contains("PATH"))
                paths = other["PATH"].split(":")
                self.test.assertThat(paths, Contains("/sbin"))
                return True

        self.assertThat(self.run_output_mock.call_count, Equals(1))
        self.run_output_mock.assert_has_calls(
            [
                mock.call(
                    [
                        "modprobe",
                        "-n",
                        "--show-depends",
                        "-d",
                        plugin.installdir,
                        "-S",
                        "4.4.2",
                        "my-fake-module",
                    ],
                    env=_check_env(self),
                )
            ]
        )

        config_file = os.path.join(plugin.builddir, ".config")
        self.assertTrue(os.path.exists(config_file))

        with open(config_file) as f:
            config_contents = f.read()

        self.assertThat(config_contents, Equals("ACCEPT=y\n"))
        self._assert_common_assets(plugin.installdir)

    @mock.patch.object(snapcraft.ProjectOptions, "kernel_arch", new="not_arm")
    def test_build_with_kconfigfile_and_firmware(self):
        self.options.kconfigfile = "config"
        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")
        self.options.kernel_initrd_firmware = [
            "lib/firmware/fake-fw-dir",
            "lib/firmware/fake-fw.bin",
        ]

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self._simulate_build(plugin.sourcedir, plugin.builddir, plugin.installdir)

        def fake_unpack(*args, **kwargs):
            modules_dir = os.path.join(
                plugin.builddir, "initrd-staging", "lib", "modules", "4.4.2"
            )
            if os.path.exists(modules_dir):
                return
            os.makedirs(modules_dir)

        self.check_call_mock.side_effect = fake_unpack

        plugin.build()

        self._assert_generic_check_call(
            plugin.builddir, plugin.installdir, plugin.os_snap
        )

        self.assertThat(self.run_mock.call_count, Equals(2))
        self.run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2", "bzImage", "modules"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "modules_install",
                        "INSTALL_MOD_PATH={}".format(plugin.installdir),
                        "firmware_install",
                        "INSTALL_FW_PATH={}".format(
                            os.path.join(plugin.installdir, "lib", "firmware")
                        ),
                    ]
                ),
            ]
        )

        config_file = os.path.join(plugin.builddir, ".config")
        self.assertTrue(os.path.exists(config_file))

        with open(config_file) as f:
            config_contents = f.read()

        self.assertThat(config_contents, Equals("ACCEPT=y\n"))
        self._assert_common_assets(plugin.installdir)
        self.assertTrue(
            os.path.exists(os.path.join(plugin.installdir, "firmware", "fake-fw-dir"))
        )

    @mock.patch.object(snapcraft.ProjectOptions, "kernel_arch", new="not_arm")
    def test_build_with_kconfigfile_and_no_firmware(self):
        self.options.kconfigfile = "config"
        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")
        self.options.kernel_with_firmware = False

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self._simulate_build(
            plugin.sourcedir,
            plugin.builddir,
            plugin.installdir,
            do_dtbs=True,
            do_firmware=False,
        )

        plugin.build()

        self._assert_generic_check_call(
            plugin.builddir, plugin.installdir, plugin.os_snap
        )

        self.assertThat(self.run_mock.call_count, Equals(2))
        self.run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2", "bzImage", "modules"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "modules_install",
                        "INSTALL_MOD_PATH={}".format(plugin.installdir),
                    ]
                ),
            ]
        )

        config_file = os.path.join(plugin.builddir, ".config")
        self.assertTrue(os.path.exists(config_file))

    @mock.patch.object(snapcraft.ProjectOptions, "kernel_arch", new="not_arm")
    def test_build_with_kconfigflavour(self):
        arch = self.project.deb_arch
        branch = "master"
        flavour = "vanilla"
        self.options.kconfigflavour = flavour

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)
        self._simulate_build(plugin.sourcedir, plugin.builddir, plugin.installdir)

        debiandir = os.path.join(plugin.sourcedir, "debian")
        os.mkdir(debiandir)
        with open(os.path.join(debiandir, "debian.env"), "w") as f:
            f.write("DEBIAN=debian.{}".format(branch))
        os.mkdir(os.path.join(plugin.sourcedir, "debian.{}".format(branch)))
        basedir = os.path.join(plugin.sourcedir, "debian.{}".format(branch), "config")
        archdir = os.path.join(
            plugin.sourcedir, "debian.{}".format(branch), "config", arch
        )
        os.mkdir(basedir)
        os.mkdir(archdir)
        commoncfg = os.path.join(basedir, "config.common.ports")
        ubuntucfg = os.path.join(basedir, "config.common.ubuntu")
        archcfg = os.path.join(archdir, "config.common.{}".format(arch))
        flavourcfg = os.path.join(archdir, "config.flavour.{}".format(flavour))

        with open(commoncfg, "w") as f:
            f.write("ACCEPT=y\n")
        with open(ubuntucfg, "w") as f:
            f.write("ACCEPT=m\n")
        with open(archcfg, "w") as f:
            f.write("ACCEPT=y\n")
        with open(flavourcfg, "w") as f:
            f.write("# ACCEPT is not set\n")

        plugin.build()

        self._assert_generic_check_call(
            plugin.builddir, plugin.installdir, plugin.os_snap
        )

        self.assertThat(self.run_mock.call_count, Equals(2))
        self.run_mock.assert_has_calls(
            [
                mock.call(["make", "-j2", "bzImage", "modules"]),
                mock.call(
                    [
                        "make",
                        "-j2",
                        "CONFIG_PREFIX={}".format(plugin.installdir),
                        "modules_install",
                        "INSTALL_MOD_PATH={}".format(plugin.installdir),
                        "firmware_install",
                        "INSTALL_FW_PATH={}".format(
                            os.path.join(plugin.installdir, "lib", "firmware")
                        ),
                    ]
                ),
            ]
        )

        config_file = os.path.join(plugin.builddir, ".config")
        self.assertTrue(os.path.exists(config_file))

        self.assertThat(
            config_file,
            FileContains(
                dedent(
                    """\
        ACCEPT=y
        ACCEPT=m
        ACCEPT=y
        # ACCEPT is not set
        """
                )
            ),
        )
        self._assert_common_assets(plugin.installdir)

    def test_build_with_missing_kernel_fails(self):
        self.options.kconfigfile = "config"

        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self._simulate_build(
            plugin.sourcedir, plugin.builddir, plugin.installdir, do_kernel=False
        )

        raised = self.assertRaises(ValueError, plugin.build)

        self.assertThat(
            str(raised),
            Equals(
                "kernel build did not output a vmlinux binary in top level "
                "dir, expected {!r}".format(
                    os.path.join(
                        plugin.builddir,
                        "arch",
                        self.project.kernel_arch,
                        "boot",
                        "bzImage",
                    )
                )
            ),
        )

    def test_build_with_missing_kernel_release_fails(self):
        self.options.kconfigfile = "config"

        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self._simulate_build(
            plugin.sourcedir, plugin.builddir, plugin.installdir, do_release=False
        )

        raised = self.assertRaises(ValueError, plugin.build)

        self.assertThat(
            str(raised),
            Equals(
                "No kernel release version info found at {!r}".format(
                    os.path.join(plugin.builddir, "include", "config", "kernel.release")
                )
            ),
        )

    def test_build_with_missing_system_map_fails(self):
        self.options.kconfigfile = "config"
        with open(self.options.kconfigfile, "w") as f:
            f.write("ACCEPT=y\n")

        plugin = kernel.KernelPlugin("test-part", self.options, self.project)

        self._simulate_build(
            plugin.sourcedir, plugin.builddir, plugin.installdir, do_system_map=False
        )

        raised = self.assertRaises(ValueError, plugin.build)

        self.assertThat(
            str(raised),
            Equals("kernel build did not output a System.map in top level dir"),
        )

    def test_enable_cross_compilation(self):
        project = snapcraft.project.Project(target_deb_arch="arm64")
        project._snap_meta = meta.snap.Snap(name="test-snap", base="core18")

        plugin = kernel.KernelPlugin("test-part", self.options, project)
        plugin.enable_cross_compilation()

        self.assertThat(
            plugin.make_cmd,
            Equals(
                [
                    "make",
                    "-j2",
                    "ARCH=arm64",
                    "CROSS_COMPILE=aarch64-linux-gnu-",
                    "PATH={}:/usr/{}/bin".format(
                        os.environ.copy().get("PATH", ""), "aarch64-linux-gnu"
                    ),
                ]
            ),
        )

    def test_override_cross_compile(self):
        project = snapcraft.project.Project(target_deb_arch="arm64")
        project._snap_meta = meta.snap.Snap(name="test-snap", base="core18")

        plugin = kernel.KernelPlugin("test-part", self.options, project)
        self.useFixture(
            fixtures.EnvironmentVariable("CROSS_COMPILE", "foo-bar-toolchain-")
        )
        plugin.enable_cross_compilation()

        self.assertThat(
            plugin.make_cmd,
            Equals(
                [
                    "make",
                    "-j2",
                    "ARCH=arm64",
                    "CROSS_COMPILE=foo-bar-toolchain-",
                    "PATH={}:/usr/{}/bin".format(
                        os.environ.copy().get("PATH", ""), "aarch64-linux-gnu"
                    ),
                ]
            ),
        )

    def test_override_cross_compile_empty(self):
        project = snapcraft.project.Project(target_deb_arch="arm64")
        project._snap_meta = meta.snap.Snap(name="test-snap", base="core18")

        plugin = kernel.KernelPlugin("test-part", self.options, project)
        self.useFixture(fixtures.EnvironmentVariable("CROSS_COMPILE", ""))
        plugin.enable_cross_compilation()

        self.assertThat(
            plugin.make_cmd,
            Equals(
                [
                    "make",
                    "-j2",
                    "ARCH=arm64",
                    "CROSS_COMPILE=aarch64-linux-gnu-",
                    "PATH={}:/usr/{}/bin".format(
                        os.environ.copy().get("PATH", ""), "aarch64-linux-gnu"
                    ),
                ]
            ),
        )

    def test_kernel_image_target_as_map(self):
        self.options.kernel_image_target = {"arm64": "Image"}
        project = snapcraft.project.Project(target_deb_arch="arm64")
        project._snap_meta = meta.snap.Snap(name="test-snap", base="core18")

        plugin = kernel.KernelPlugin("test-part", self.options, project)

        self.assertThat(plugin.make_targets, Equals(["Image", "modules", "dtbs"]))

    def test_kernel_image_target_as_string(self):
        self.options.kernel_image_target = "Image"
        project = snapcraft.project.Project(target_deb_arch="arm64")
        project._snap_meta = meta.snap.Snap(name="test-snap", base="core18")

        plugin = kernel.KernelPlugin("test-part", self.options, project)

        self.assertThat(plugin.make_targets, Equals(["Image", "modules", "dtbs"]))

    def test_kernel_image_target_non_existent(self):
        class Options:
            build_parameters = []
            kconfigfile = None
            kdefconfig = []
            kconfigs = []
            kernel_with_firmware = True
            kernel_initrd_modules = []
            kernel_initrd_firmware = []
            kernel_device_trees = []
            kernel_initrd_compression = "gz"

        project = snapcraft.project.Project(target_deb_arch="arm64")
        project._snap_meta = meta.snap.Snap(name="test-snap", base="core18")

        plugin = kernel.KernelPlugin("test-part", self.options, project)

        self.assertThat(plugin.make_targets, Equals(["bzImage", "modules", "dtbs"]))

    @mock.patch.object(storeapi.StoreClient, "download")
    def test_pull(self, download_mock):
        plugin = kernel.KernelPlugin("test-part", self.options, self.project)
        plugin.pull()

        download_mock.assert_called_once_with(
            "core",
            risk="stable",
            track=None,
            download_path=plugin.os_snap,
            arch=self.project.deb_arch,
            except_hash="",
        )

    def test_unsupported_base(self):
        self.project._snap_meta.base = "unsupported-base"

        raised = self.assertRaises(
            errors.PluginBaseError,
            kernel.KernelPlugin,
            "test-part",
            self.options,
            self.project,
        )

        self.assertThat(raised.part_name, Equals("test-part"))
        self.assertThat(raised.base, Equals("unsupported-base"))


@pytest.mark.parametrize(
    "deb_arch,expected_target",
    [
        ("amd64", "bzImage"),
        ("i386", "bzImage"),
        ("arm64", "Image.gz"),
        ("armhf", "zImage"),
    ],
)
def test_target(deb_arch, expected_target):
    class Options:
        build_parameters = []
        kconfigfile = None
        kdefconfig = []
        kconfigs = []
        kernel_image_target = ""
        kernel_with_firmware = True
        kernel_initrd_modules = []
        kernel_initrd_firmware = []
        kernel_device_trees = []
        kernel_initrd_compression = "gz"

    project = snapcraft.project.Project(target_deb_arch=deb_arch)
    project._snap_meta = meta.snap.Snap(name="test-snap", base="core18")

    plugin = kernel.KernelPlugin("test-part", Options(), project)

    assert plugin.kernel_image_target == expected_target



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import pathlib
import subprocess

import fixtures

from snapcraft.internal import review_tools
from tests import unit


class RunTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.review_tools_path = "/snap/bin/review-tools.snap-review"
        self.fake_check_output = fixtures.MockPatch(
            "subprocess.check_output", return_value=b""
        )
        self.useFixture(self.fake_check_output)

        self.user_common_path = pathlib.Path(self.path) / "common"
        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.review_tools._runner._get_review_tools_user_common",
                return_value=self.user_common_path,
            )
        )

        self.fake_snap_path = pathlib.Path("fake.snap")
        self.fake_snap_path.touch()

    def assert_fake_check_output_called(self):
        self.fake_check_output.mock.assert_called_once_with(
            [
                self.review_tools_path,
                self.user_common_path / self.fake_snap_path,
                "--json",
                "--allow-classic",
            ],
            env={"SNAP_ENFORCE_RESQUASHFS": "0"},
            stderr=subprocess.STDOUT,
        )

    def test_review_good(self):
        review_tools.run(snap_filename="fake.snap")

        self.assert_fake_check_output_called()

    def test_review_warnings(self):
        self.fake_check_output.mock.side_effect = subprocess.CalledProcessError(
            cmd=[self.review_tools_path, "fake.snap", "--json"],
            returncode=2,
            output=json.dumps(
                {
                    "snap.v2_functional": {"error": {}, "warn": {}},
                    "snap.v2_security": {
                        "warn": {
                            "security-snap-v2:security_issue": {
                                "text": "(NEEDS REVIEW) security message."
                            }
                        },
                        "error": {},
                    },
                    "snap.v2_lint": {"error": {}, "warn": {}},
                }
            ).encode(),
        )

        self.assertRaises(
            review_tools.errors.ReviewError, review_tools.run, snap_filename="fake.snap"
        )
        self.assert_fake_check_output_called()

    def test_review_errors(self):
        self.fake_check_output.mock.side_effect = subprocess.CalledProcessError(
            cmd=[self.review_tools_path, "fake.snap", "--json", "--allow-classic"],
            returncode=3,
            output=json.dumps(
                {
                    "snap.v2_functional": {"error": {}, "warn": {}},
                    "snap.v2_security": {
                        "error": {
                            "security-snap-v2:security_issue": {
                                "text": "(NEEDS REVIEW) security message."
                            }
                        },
                        "warn": {},
                    },
                    "snap.v2_lint": {"error": {}, "warn": {}},
                }
            ).encode(),
        )

        self.assertRaises(
            review_tools.errors.ReviewError, review_tools.run, snap_filename="fake.snap"
        )
        self.assert_fake_check_output_called()

    def test_review_unkown_error_bubbles_up(self):
        self.fake_check_output.mock.side_effect = subprocess.CalledProcessError(
            cmd=[self.review_tools_path, "fake.snap", "--json", "--allow-classic"],
            returncode=4,
            output=b"unknown",
        )

        self.assertRaises(
            subprocess.CalledProcessError, review_tools.run, snap_filename="fake.snap"
        )
        self.assert_fake_check_output_called()

    def test_review_tools_missing(self):
        self.fake_check_output.mock.side_effect = FileNotFoundError(
            "review-tools.snap-review"
        )

        self.assertRaises(
            review_tools.errors.ReviewToolMissing,
            review_tools.run,
            snap_filename="fake.snap",
        )
        self.assert_fake_check_output_called()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The python plugin can be used for 3 based parts.

It can be used for python projects where you would want to do:

    - import python modules with a requirements.txt
    - build a python project that has a setup.py
    - install packages straight from pip

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - requirements
      (list of strings)
      List of paths to requirements files.

    - constraints
      (list of strings)
      List of paths to constraint files.

    - python-packages
      (list)
      A list of dependencies to get from PyPI. If needed, pip,
      setuptools and wheel can be upgraded here.

This plugin also interprets these specific build-environment entries:

    - SNAPCRAFT_PYTHON_INTERPRETER
      (default: python3)
      The interpreter binary to search for in PATH.

    - SNAPCRAFT_PYTHON_VENV_ARGS
      Additional arguments for venv.

By default this plugin uses python from the base, if a part using
this plugin uses a build-base other than that of the base, or a
different interpreter is desired, it must be bundled in the snap
(including venv) and must be in PATH.

It is required to bundle python when creating a snap that uses
classic confinement, this can be accomplished on Ubuntu by
adding stage-packages (i.e.; python3-venv).

Use of python3-<python-package> in stage-packages will force the
inclusion of the python interpreter.
"""

import shlex
from textwrap import dedent
from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import PluginV2


class PythonPlugin(PluginV2):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "requirements": {
                    "type": "array",
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": [],
                },
                "constraints": {
                    "type": "array",
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": [],
                },
                "python-packages": {
                    "type": "array",
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": ["pip", "setuptools", "wheel"],
                },
            },
        }

    def get_build_snaps(self) -> Set[str]:
        return set()

    def get_build_packages(self) -> Set[str]:
        return {"findutils", "python3-dev", "python3-venv"}

    def get_build_environment(self) -> Dict[str, str]:
        return {
            # Add PATH to the python interpreter we always intend to use with
            # this plugin. It can be user overridden, but that is an explicit
            # choice made by a user.
            "PATH": "${SNAPCRAFT_PART_INSTALL}/bin:${PATH}",
            "SNAPCRAFT_PYTHON_INTERPRETER": "python3",
            "SNAPCRAFT_PYTHON_VENV_ARGS": "",
        }

    def get_build_commands(self) -> List[str]:
        # $SNAPCRAFT_PYTHON_INTERPRETER (as -m venv will create the link to the
        # interpreter invoked).
        build_commands = [
            '"${SNAPCRAFT_PYTHON_INTERPRETER}" -m venv ${SNAPCRAFT_PYTHON_VENV_ARGS} "${SNAPCRAFT_PART_INSTALL}"',
            'SNAPCRAFT_PYTHON_VENV_INTERP_PATH="${SNAPCRAFT_PART_INSTALL}/bin/${SNAPCRAFT_PYTHON_INTERPRETER}"',
        ]

        if self.options.constraints:
            constraints = " ".join(f"-c {c!r}" for c in self.options.constraints)
        else:
            constraints = ""

        if self.options.python_packages:
            python_packages = " ".join(
                [shlex.quote(pkg) for pkg in self.options.python_packages]
            )
            python_packages_cmd = f"pip install {constraints} -U {python_packages}"
            build_commands.append(python_packages_cmd)

        if self.options.requirements:
            requirements = " ".join(f"-r {r!r}" for r in self.options.requirements)
            requirements_cmd = f"pip install {constraints} -U {requirements}"
            build_commands.append(requirements_cmd)

        build_commands.append(f"[ -f setup.py ] && pip install {constraints} -U .")

        # Now fix shebangs.
        # TODO: replace with snapcraftctl once the two scripts are consolidated
        # and use mangling.rewrite_python_shebangs.
        build_commands.append(
            dedent(
                """\
            find "${SNAPCRAFT_PART_INSTALL}" -type f -executable -print0 | xargs -0 \
                sed -i "1 s|^#\\!${SNAPCRAFT_PYTHON_VENV_INTERP_PATH}.*$|#\\!/usr/bin/env ${SNAPCRAFT_PYTHON_INTERPRETER}|"
            """
            )
        )

        # Lastly, fix the symlink to the "real" python3 interpreter.
        # TODO: replace with snapcraftctl (create_relative_symlinks).
        build_commands.append(
            dedent(
                """\
            determine_link_target() {
                opts_state="$(set +o +x | grep xtrace)"
                interp_dir="$(dirname "${SNAPCRAFT_PYTHON_VENV_INTERP_PATH}")"
                # Determine python based on PATH, then resolve it, e.g:
                # (1) /home/ubuntu/.venv/snapcraft/bin/python3 -> /usr/bin/python3.8
                # (2) /usr/bin/python3 -> /usr/bin/python3.8
                # (3) /root/stage/python3 -> /root/stage/python3.8
                # (4) /root/parts/<part>/install/usr/bin/python3 -> /root/parts/<part>/install/usr/bin/python3.8
                python_path="$(which "${SNAPCRAFT_PYTHON_INTERPRETER}")"
                python_path="$(readlink -e "${python_path}")"
                for dir in "${SNAPCRAFT_PART_INSTALL}" "${SNAPCRAFT_STAGE}"; do
                    if  echo "${python_path}" | grep -q "${dir}"; then
                        python_path="$(realpath --strip --relative-to="${interp_dir}" \\
                                "${python_path}")"
                        break
                    fi
                done
                echo "${python_path}"
                eval "${opts_state}"
            }

            python_path="$(determine_link_target)"
            ln -sf "${python_path}" "${SNAPCRAFT_PYTHON_VENV_INTERP_PATH}"
            """
            )
        )

        return build_commands



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import logging
import os
import shlex
from subprocess import CalledProcessError
from typing import List

from snapcraft.project import Project
from snapcraft.internal import common, errors
from snapcraft.internal.meta.package_repository import PackageRepository

logger = logging.getLogger(__name__)


class PluginV1:
    @classmethod
    def schema(cls):
        """Return a json-schema for the plugin's properties as a dictionary.
        Of importance to plugin authors is the 'properties' keyword and
        optionally the 'requires' keyword with a list of required
        'properties'.

        By default the properties will be that of a standard VCS,
        override in custom implementations if required.
        """
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {},
        }

    @classmethod
    def get_pull_properties(cls):
        return []

    @classmethod
    def get_build_properties(cls):
        return []

    @classmethod
    def get_required_package_repositories(self) -> List[PackageRepository]:
        """Define additional deb source lines using templates variables."""
        return list()

    @property
    def stage_packages(self):
        return self._stage_packages

    @stage_packages.setter
    def stage_packages(self, value):
        self._stage_packages = value

    def __init__(self, name, options, project=None):
        self.name = name
        self.build_snaps = []
        self.stage_snaps = []
        self.build_packages = []
        self._stage_packages = []

        with contextlib.suppress(AttributeError):
            self._stage_packages = options.stage_packages.copy()
        with contextlib.suppress(AttributeError):
            self.build_packages = options.build_packages.copy()
        with contextlib.suppress(AttributeError):
            self.build_snaps = options.build_snaps.copy()
        with contextlib.suppress(AttributeError):
            self.stage_snaps = options.stage_snaps.copy()

        self.project = project
        self.options = options

        if project:
            if isinstance(project, Project) and project._get_build_base() not in (
                "core",
                "core16",
                "core18",
            ):
                raise errors.PluginBaseError(
                    part_name=self.name, base=project._get_build_base()
                )

            self.partdir = os.path.join(project.parts_dir, name)
        else:
            self.partdir = os.path.join(os.getcwd(), "parts", name)

        self.sourcedir = os.path.join(self.partdir, "src")
        self.installdir = os.path.join(self.partdir, "install")
        self.statedir = os.path.join(self.partdir, "state")

        self.build_basedir = os.path.join(self.partdir, "build")
        source_subdir = getattr(self.options, "source_subdir", None)
        if source_subdir:
            self.builddir = os.path.join(self.build_basedir, source_subdir)
        else:
            self.builddir = self.build_basedir

        # By default, snapcraft does an in-source build. Set this property to
        # True if that's not desired.
        self.out_of_source_build = False

    # The API
    def pull(self):
        """Pull the source code and/or internal prereqs to build the part."""
        pass

    def clean_pull(self):
        """Clean the pulled source for this part."""
        pass

    def build(self):
        """Build the source code retrieved from the pull phase."""
        pass

    def clean_build(self):
        """Clean the artifacts that resulted from building this part."""
        pass

    def get_manifest(self):
        """Return the information to record after the build of this part.

        :rtype: dict
        """
        pass

    def snap_fileset(self):
        """Return a list of files to include or exclude in the resulting snap

        The staging phase of a plugin's lifecycle may populate many things
        into the staging directory in order to succeed in building a
        project.
        During the stripping phase and in order to have a clean snap, the
        plugin can provide additional logic for stripping build components
        from the final snap and alleviate the part author from doing so for
        repetetive filesets.

        These are the rules to honor when creating such list:

            - includes can be just listed
            - excludes must be preceded by -

        For example::
            (['bin', 'lib', '-include'])
        """
        return []

    def env(self, root):
        """Return a list with the execution environment for building.

        Plugins often need special environment variables exported to the
        system for some builds to take place. This is a list of strings
        of the form key=value. The parameter root is the path to this part.

        :param str root: The root for the part
        """
        return []

    def enable_cross_compilation(self):
        """Enable cross compilation for the plugin."""
        raise errors.CrossCompilationNotSupported(part_name=self.name)

    @property
    def parallel_build_count(self):
        """Number of CPU's to use for building.

        Number comes from `project.parallel_build_count` unless the part
        has defined `disable-parallel` as `True`.
        """
        if getattr(self.options, "disable_parallel", False):
            return 1
        else:
            return self.project.parallel_build_count

    # Helpers
    def run(self, cmd, cwd=None, **kwargs):
        if not cwd:
            cwd = self.builddir
        cmd_string = " ".join([shlex.quote(c) for c in cmd])
        print(cmd_string)
        os.makedirs(cwd, exist_ok=True)
        try:
            return common.run(cmd, cwd=cwd, **kwargs)
        except CalledProcessError as process_error:
            raise errors.SnapcraftPluginCommandError(
                command=cmd, part_name=self.name, exit_code=process_error.returncode
            ) from process_error

    def run_output(self, cmd, cwd=None, **kwargs):
        if not cwd:
            cwd = self.builddir
        os.makedirs(cwd, exist_ok=True)
        try:
            return common.run_output(cmd, cwd=cwd, **kwargs)
        except CalledProcessError as process_error:
            raise errors.SnapcraftPluginCommandError(
                command=cmd, part_name=self.name, exit_code=process_error.returncode
            ) from process_error



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import textwrap

import yaml as yaml_utils
from testtools.matchers import Contains, Equals

from tests import integration


class SnapcraftctlSetVersionTestCase(integration.TestCase):
    def test_set_version(self):
        self.construct_yaml(
            version=None,
            adopt_info="my-part",
            parts=textwrap.dedent(
                """\
                my-part:
                  plugin: nil
                  override-pull: snapcraftctl set-version override-version
                """
            ),
        )

        self.run_snapcraft("prime")

        with open(os.path.join("prime", "meta", "snap.yaml")) as f:
            y = yaml_utils.load(f)

        self.assertThat(y["version"], Equals("override-version"))

    def test_set_version_no_overwrite(self):
        self.construct_yaml(
            version="test-version",
            adopt_info="my-part",
            parts=textwrap.dedent(
                """\
                my-part:
                  plugin: nil
                  override-pull: snapcraftctl set-version override-version
                """
            ),
        )

        self.run_snapcraft("prime")

        with open(os.path.join("prime", "meta", "snap.yaml")) as f:
            y = yaml_utils.load(f)

        self.assertThat(y["version"], Equals("test-version"))

    def test_set_version_twice_errors(self):
        self.construct_yaml(
            version=None,
            adopt_info="my-part",
            parts=textwrap.dedent(
                """\
                my-part:
                  plugin: nil
                  override-pull: snapcraftctl set-version override-version
                  override-prime: snapcraftctl set-version no-this-version
                """
            ),
        )

        raised = self.assertRaises(
            subprocess.CalledProcessError, self.run_snapcraft, "prime"
        )
        self.assertThat(
            raised.output,
            Contains("Unable to set version: it was already set in the 'pull' step"),
        )
        self.assertThat(raised.output, Contains("Failed to run 'override-prime'"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
import base64
import io
import os
import pathlib
from typing import Optional, TextIO

import configparser

from . import errors


class Config(abc.ABC):
    def __init__(self) -> None:
        self.parser = configparser.ConfigParser()
        self.load()

    @abc.abstractmethod
    def _get_section_name(self) -> str:
        """Return section name."""

    @abc.abstractmethod
    def _get_config_path(self) -> pathlib.Path:
        """Return Path to configuration file."""

    def get(
        self, option_name: str, section_name: Optional[str] = None
    ) -> Optional[str]:
        """Return content of section_name/option_name or None if not found."""
        if section_name is None:
            section_name = self._get_section_name()
        try:
            return self.parser.get(section_name, option_name)
        except (configparser.NoSectionError, configparser.NoOptionError, KeyError):
            return None

    def set(
        self, option_name: str, value: str, section_name: Optional[str] = None
    ) -> None:
        """Set value to section_name/option_name."""
        if not section_name:
            section_name = self._get_section_name()
        if not self.parser.has_section(section_name):
            self.parser.add_section(section_name)
        self.parser.set(section_name, option_name, value)

    def is_section_empty(self, section_name: Optional[str] = None) -> bool:
        """Check if section_name is empty."""
        if section_name is None:
            section_name = self._get_section_name()

        if self.parser.has_section(section_name):
            if self.parser.options(section_name):
                return False
        return True

    def _load_potentially_base64_config(self, config_content: str) -> None:
        try:
            self.parser.read_string(config_content)
        except configparser.Error as parser_error:
            # The config may be base64-encoded, try decoding it
            try:
                decoded_config_content = base64.b64decode(config_content).decode()
            except base64.binascii.Error:  # type: ignore
                # It wasn't base64, so use the original error
                raise errors.InvalidLoginConfig(parser_error)

            try:
                self.parser.read_string(decoded_config_content)
            except configparser.Error as parser_error:
                raise errors.InvalidLoginConfig(parser_error)

    def load(self, *, config_fd: TextIO = None) -> None:
        if config_fd is not None:
            config_content = config_fd.read()
        elif self._get_config_path().exists():
            with self._get_config_path().open() as config_file:
                config_content = config_file.read()
        else:
            return

        self._load_potentially_base64_config(config_content)

    def save(self, *, config_fd: Optional[TextIO] = None, encode: bool = False) -> None:
        with io.StringIO() as config_buffer:
            self.parser.write(config_buffer)
            config_content = config_buffer.getvalue()
            if encode:
                config_content = base64.b64encode(config_content.encode()).decode()

            if config_fd:
                print(config_content, file=config_fd)
            else:
                with self._get_config_path().open("w") as config_file:
                    print(config_content, file=config_file)
                    config_file.flush()
                    os.fsync(config_file.fileno())

    def clear(self, section_name: Optional[str] = None) -> None:
        if section_name is None:
            section_name = self._get_section_name()

        self.parser.remove_section(self._get_section_name())



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016, 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import http.server
import json
import logging
import os
import urllib.parse
from collections import OrderedDict
from datetime import datetime

import pymacaroons

# we do not want snapcraft imports for the integration tests
try:
    from snapcraft import yaml_utils
except ImportError:
    import yaml as yaml_utils  # type: ignore

logger = logging.getLogger(__name__)


class BaseHTTPRequestHandler(http.server.BaseHTTPRequestHandler):
    def log_message(*args):
        logger.debug(args)

    def raise_not_implemented(self, path):
        logger.error(
            "Not implemented {} in server: {}".format(path, self.__class__.__name__)
        )
        raise NotImplementedError(path)


class FakeFileHTTPRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        if self.path.endswith("404-not-found"):
            self.send_response(404)
            self.end_headers()
        else:
            data = "Test fake file"
            self.send_response(200)
            self.send_header("Content-Length", len(data))
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(data.encode())


class FakePartsServer(http.server.HTTPServer):
    def __init__(self, server_address):
        super().__init__(server_address, FakePartsRequestHandler)


class FakePartsRequestHandler(BaseHTTPRequestHandler):

    _date_format = "%a, %d %b %Y %H:%M:%S GMT"
    _parts_date = datetime(2016, 7, 7, 10, 0, 20)

    def do_GET(self):
        logger.debug("Handling getting parts")
        if "If-Modified-Since" in self.headers:
            ims_date = datetime.strptime(
                self.headers["If-Modified-Since"], self._date_format
            )
        else:
            ims_date = None
        if ims_date is not None and ims_date >= self._parts_date:
            self.send_response(304)
            response = {}
        elif "CUSTOM_PARTS" in os.environ:
            self.send_response(200)
            response = OrderedDict(
                (
                    (
                        "curl-custom",
                        OrderedDict(
                            (
                                ("plugin", "autotools"),
                                ("source", "http://curl.org"),
                                ("description", "custom curl part"),
                                ("maintainer", "none"),
                            )
                        ),
                    ),
                )
            )
        else:
            self.send_response(200)
            response = OrderedDict(
                (
                    (
                        "curl",
                        OrderedDict(
                            (
                                ("plugin", "autotools"),
                                ("source", "http://curl.org"),
                                ("description", "test entry for curl"),
                                ("maintainer", "none"),
                            )
                        ),
                    ),
                    (
                        "part1",
                        OrderedDict(
                            (
                                ("plugin", "go"),
                                ("source", "http://source.tar.gz"),
                                ("description", "test entry for part1"),
                                ("maintainer", "none"),
                            )
                        ),
                    ),
                    (
                        "long-described-part",
                        OrderedDict(
                            (
                                ("plugin", "go"),
                                ("source", "http://source.tar.gz"),
                                (
                                    "description",
                                    "this is a repetitive description " * 3,
                                ),
                                ("maintainer", "none"),
                            )
                        ),
                    ),
                    (
                        "multiline-part",
                        OrderedDict(
                            (
                                ("plugin", "go"),
                                ("source", "http://source.tar.gz"),
                                (
                                    "description",
                                    "this is a multiline description\n" * 3,
                                ),
                                ("maintainer", "none"),
                            )
                        ),
                    ),
                )
            )
        self.send_header("Content-Type", "text/plain")
        if "NO_CONTENT_LENGTH" not in os.environ:
            self.send_header("Content-Length", "1000")
        self.send_header("Last-Modified", self._parts_date.strftime(self._date_format))
        self.send_header("ETag", "1111")
        self.end_headers()
        self.wfile.write(yaml_utils.dump(response).encode())


class FakePartsWikiServer(http.server.HTTPServer):
    def __init__(self, server_address):
        super().__init__(server_address, FakePartsWikiRequestHandler)


class FakePartsWikiRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        logger.debug("Handling getting parts")
        self.send_response(200)
        response = """
---
origin: https://github.com/sergiusens/curl.git
parts: [curl]
description:
  Description here
maintainer: none
"""
        self.send_header("Content-Type", "text/plain")
        if "NO_CONTENT_LENGTH" not in os.environ:
            self.send_header("Content-Length", len(response.encode()))
        self.end_headers()
        self.wfile.write(response.encode())


class FakePartsWikiWithSlashesServer(http.server.HTTPServer):
    def __init__(self, server_address):
        super().__init__(server_address, FakePartsWikiWithSlashesRequestHandler)


class FakePartsWikiWithSlashesRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        logger.debug("Handling getting parts")
        self.send_response(200)
        response = """
---
origin: https://github.com/sergiusens/curl.git
parts: [curl/a]
description:
  Description here
maintainer: none
---
origin: https://github.com/sergiusens/curl.git
parts: [curl-a]
description:
  Description here
maintainer: none
"""
        self.send_header("Content-Type", "text/plain")
        if "NO_CONTENT_LENGTH" not in os.environ:
            self.send_header("Content-Length", len(response.encode()))
        self.end_headers()
        self.wfile.write(response.encode())


class FakePartsWikiOriginServer(http.server.HTTPServer):
    def __init__(self, server_address):
        super().__init__(server_address, FakePartsWikiOriginRequestHandler)


class FakePartsWikiOriginRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        logger.debug("Handling getting part origin")
        self.send_response(200)
        response = """
parts:
  somepart:
    source: https://github.com/someuser/somepart.git
    plugin: nil
"""
        self.send_header("Content-Type", "text/plain")
        if "NO_CONTENT_LENGTH" not in os.environ:
            self.send_header("Content-Length", len(response.encode()))
        self.end_headers()
        self.wfile.write(response.encode())


class FakeSSOServer(http.server.HTTPServer):
    def __init__(self, fake_store, server_address):
        super().__init__(server_address, FakeSSORequestHandler)
        self.fake_store = fake_store


class FakeSSORequestHandler(BaseHTTPRequestHandler):

    _API_PATH = "/api/v2/"

    def do_POST(self):
        parsed_path = urllib.parse.urlparse(self.path)
        tokens_discharge_path = urllib.parse.urljoin(self._API_PATH, "tokens/discharge")
        tokens_refresh_path = urllib.parse.urljoin(self._API_PATH, "tokens/refresh")
        if parsed_path.path.startswith(tokens_discharge_path):
            self._handle_tokens_discharge_request()
        elif parsed_path.path.startswith(tokens_refresh_path):
            self._handle_tokens_refresh_request()
        else:
            self.raise_not_implemented(self.path)

    def _handle_tokens_discharge_request(self):
        string_data = self.rfile.read(int(self.headers["Content-Length"])).decode(
            "utf8"
        )
        data = json.loads(string_data)
        logger.debug("Handling tokens discharge request with content {}".format(data))
        if data["password"] == "test correct password" and (
            "otp" not in data or data["otp"] == "test correct one-time password"
        ):
            self._send_tokens_discharge()
        elif data["password"] == "test requires 2fa":
            self._send_twofactor_required_error()
        elif data["password"] == "test 401 invalid json":
            self._send_401_invalid_json()
        else:
            self._send_invalid_credentials_error()

    def _send_tokens_discharge(self):
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        response = {"discharge_macaroon": pymacaroons.Macaroon().serialize()}
        self.wfile.write(json.dumps(response).encode())

    def _send_twofactor_required_error(self):
        self.send_response(401)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        response = {
            "error_list": [
                {
                    "code": "twofactor-required",
                    "message": "2-factor authentication required.",
                }
            ]
        }
        self.wfile.write(json.dumps(response).encode())

    def _send_401_invalid_json(self):
        self.send_response(401)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(b"invalid{json")

    def _send_invalid_credentials_error(self):
        self.send_response(401)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        response = {
            "error_list": [
                {
                    "code": "invalid-credentials",
                    "message": "Provided email/password is not correct.",
                }
            ]
        }
        self.wfile.write(json.dumps(response).encode())

    def _handle_tokens_refresh_request(self):
        self._send_tokens_discharge()
        self.server.fake_store.needs_refresh = False



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import pathlib
import re
import shutil
import subprocess
import sys
from typing import Dict, Set

from snapcraft.internal import errors, repo

logger = logging.getLogger(__name__)


class RosdepPackageNotFoundError(errors.SnapcraftError):
    fmt = "rosdep cannot find Catkin package {package!r}"

    def __init__(self, package):
        super().__init__(package=package)


class RosdepDependencyNotResolvedError(errors.SnapcraftError):
    fmt = "rosdep cannot resolve {dependency!r} into a valid dependency"

    def __init__(self, dependency):
        super().__init__(dependency=dependency)


class RosdepUnexpectedResultError(errors.SnapcraftError):
    fmt = (
        "Received unexpected result from rosdep when trying to resolve "
        "{dependency!r}:\n{output}"
    )

    def __init__(self, dependency, output):
        super().__init__(dependency=dependency, output=output)


class RosdepInitializationError(errors.SnapcraftError):
    fmt = "Failed to initialize rosdep: {message}"

    def __init__(self, message):
        super().__init__(message=message)


def _parse_rosdep_resolve_dependencies(
    dependency_name: str, output: str
) -> Dict[str, Set[str]]:
    # The output of rosdep follows the pattern:
    #
    #    #apt
    #    package1
    #    package2
    #    #pip
    #    pip-package1
    #    pip-package2
    #
    # Split these out into a dict of dependency type -> dependencies.
    delimiters = re.compile(r"\n|\s")
    lines = delimiters.split(output)
    dependencies: Dict[str, Set[str]] = {}
    dependency_set = None
    for line in lines:
        line = line.strip()
        if line.startswith("#"):
            key = line.strip("# ")
            dependencies[key] = set()
            dependency_set = dependencies[key]
        elif line:
            if dependency_set is None:
                raise RosdepUnexpectedResultError(dependency_name, output)
            else:
                dependency_set.add(line)

    return dependencies


class Rosdep:
    def __init__(
        self,
        *,
        ros_distro,
        ros_version,
        ros_package_path,
        rosdep_path,
        ubuntu_distro,
        base,
        target_arch,
    ):
        self._ros_distro = ros_distro
        self._ros_version = ros_version
        self._ros_package_path = ros_package_path
        self._rosdep_path = rosdep_path
        self._ubuntu_distro = ubuntu_distro
        self._base = base

        self._rosdep_stage_packages_path = (
            pathlib.Path(self._rosdep_path) / "stage_packages"
        )
        self._rosdep_install_path = os.path.join(self._rosdep_path, "install")
        self._rosdep_sources_path = os.path.join(self._rosdep_path, "sources.list.d")
        self._rosdep_cache_path = os.path.join(self._rosdep_path, "cache")
        self._target_arch = target_arch

    def setup(self):
        # Make sure we can run multiple times without error, while leaving the
        # capability to re-initialize, by making sure we clear the sources.
        if os.path.exists(self._rosdep_sources_path):
            shutil.rmtree(self._rosdep_sources_path)

        os.makedirs(self._rosdep_sources_path)
        os.makedirs(self._rosdep_install_path, exist_ok=True)
        os.makedirs(self._rosdep_cache_path, exist_ok=True)
        self._rosdep_stage_packages_path.mkdir(exist_ok=True)

        # rosdep isn't necessarily a dependency of the project, so we'll unpack
        # it off to the side and use it from there.
        logger.info("Installing rosdep...")
        repo.Ubuntu.fetch_stage_packages(
            package_names=["python-rosdep"],
            stage_packages_path=self._rosdep_stage_packages_path,
            base=self._base,
            target_arch=self._target_arch,
        )
        repo.Ubuntu.unpack_stage_packages(
            stage_packages_path=self._rosdep_stage_packages_path,
            install_path=pathlib.Path(self._rosdep_install_path),
        )

        logger.info("Initializing rosdep database...")
        try:
            self._run(["init"])
        except subprocess.CalledProcessError as e:
            output = e.output.decode(sys.getfilesystemencoding()).strip()
            raise RosdepInitializationError(
                "Error initializing rosdep database:\n{}".format(output)
            )

        logger.info("Updating rosdep database...")
        try:
            self._run(["update", "--include-eol-distros"])
        except subprocess.CalledProcessError as e:
            output = e.output.decode(sys.getfilesystemencoding()).strip()
            raise RosdepInitializationError(
                "Error updating rosdep database:\n{}".format(output)
            )

    def get_dependencies(self, package_name=None):
        """Obtain dependencies for a given package, or entire workspace.

        :param str package_name: Package name for which dependences will be
                                 obtained. If not provided, will obtain
                                 dependencies for the entire workspace.
        """
        command = ["keys"]
        if package_name:
            command.append(package_name)
        else:
            # Adding a few flags that will make rosdep search the entire
            # workspace:
            #
            # -a: select all packages in workspace
            # -i: ignore any resolved keys that are satisfied by other packages
            #     in the workspace
            command.append("-a")
            command.append("-i")
        try:
            output = self._run(command).strip()
            if output:
                return set(output.split("\n"))
            else:
                return set()
        except subprocess.CalledProcessError:
            raise RosdepPackageNotFoundError(package_name)

    def resolve_dependency(self, dependency_name):
        try:
            # rosdep needs three pieces of information here:
            #
            # 1) The dependency we're trying to lookup.
            # 2) The rosdistro being used.
            # 3) The version of Ubuntu being used, even if we're running on
            #    something else.
            output = self._run(
                [
                    "resolve",
                    dependency_name,
                    "--rosdistro",
                    self._ros_distro,
                    "--os",
                    "ubuntu:{}".format(self._ubuntu_distro),
                ]
            )
        except subprocess.CalledProcessError:
            raise RosdepDependencyNotResolvedError(dependency_name)

        return _parse_rosdep_resolve_dependencies(dependency_name, output)

    def _run(self, arguments):
        env = os.environ.copy()

        # Use our own private rosdep and its python
        env["PATH"] = os.path.join(self._rosdep_install_path, "usr", "bin")
        env["PYTHONPATH"] = os.path.join(
            self._rosdep_install_path, "usr", "lib", "python2.7", "dist-packages"
        )

        if self._ros_version == "2":
            env["ROS_PYTHON_VERSION"] = "3"
        else:
            env["ROS_PYTHON_VERSION"] = "2"

        # By default, rosdep uses /etc/ros/rosdep to hold its sources list. We
        # don't want that here since we don't want to touch the host machine
        # (not to mention it would require sudo), so we can redirect it via
        # this environment variable
        env["ROSDEP_SOURCE_PATH"] = self._rosdep_sources_path

        # By default, rosdep saves its cache in $HOME/.ros, which we shouldn't
        # access here, so we'll redirect it with this environment variable.
        env["ROS_HOME"] = self._rosdep_cache_path

        env["ROS_VERSION"] = self._ros_version

        # This environment variable tells rosdep which directory to recursively
        # search for packages.
        env["ROS_PACKAGE_PATH"] = self._ros_package_path

        return (
            subprocess.check_output(["rosdep"] + arguments, env=env)
            .decode("utf8")
            .strip()
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

from testtools.matchers import Contains, Equals, FileExists

from snapcraft.internal import errors

from . import CommandBaseNoFifoTestCase, CommandBaseTestCase


class SetGradeCommandTestCase(CommandBaseTestCase):
    def test_set_grade(self):
        for grade in ["stable", "devel"]:
            self.run_command(["set-grade", grade])
            self.assertThat(self.call_fifo, FileExists())

            with open(self.call_fifo, "r") as f:
                data = json.loads(f.read())

            self.assertThat(data, Contains("function"))
            self.assertThat(data, Contains("args"))

            self.assertThat(data["function"], Equals("set-grade"))
            self.assertThat(data["args"], Equals({"grade": grade}))

    def test_invalid_grades(self):
        for grade in ["", "invalid-grade"]:
            result = self.run_command(["set-grade", ""])
            assert result != 0

    def test_set_grade_error(self):
        # If there is a string in the feedback, it should be considered an
        # error
        with open(self.feedback_fifo, "w") as f:
            f.write("this is an error\n")

        assert self.run_command(["set-grade", "stable"]).exit_code == -1


class SetGradeCommandWithoutFifoTestCase(CommandBaseNoFifoTestCase):
    def test_set_grade_without_fifo(self):
        raised = self.assertRaises(
            errors.SnapcraftEnvironmentError, self.run_command, ["set-grade", "stable"],
        )

        self.assertThat(
            str(raised),
            Contains(
                "environment variable must be defined. Note that this utility is "
                "only designed for use within a snapcraft.yaml"
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
import pathlib
import tempfile
from textwrap import dedent
from unittest import mock

import fixtures
from testtools.matchers import (
    Contains,
    Equals,
    FileExists,
    HasLength,
    Is,
    IsInstance,
    Not,
)

import tests
from snapcraft import storeapi
from snapcraft.storeapi import errors, http_clients
from snapcraft.storeapi.v2 import channel_map, releases, whoami, validation_sets
from tests import fixture_setup, unit


class StoreTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.fake_store = self.useFixture(fixture_setup.FakeStore())
        self.client = storeapi.StoreClient()


class LoginTestCase(StoreTestCase):
    def test_login_successful(self):
        self.client.login(email="dummy email", password="test correct password")
        self.assertIsNotNone(self.client.auth_client.auth)

    def test_login_successful_with_one_time_password(self):
        self.client.login(
            email="dummy email",
            password="test correct password",
            otp="test correct one-time password",
        )
        self.assertIsNotNone(self.client.auth_client.auth)

    def test_login_successful_with_package_attenuation(self):
        self.client.login(
            email="dummy email",
            password="test correct password",
            packages=[{"name": "foo", "series": "16"}],
        )
        self.assertIsNotNone(self.client.auth_client.auth)

    def test_login_successful_with_channel_attenuation(self):
        self.client.login(
            email="dummy email", password="test correct password", channels=["edge"]
        )
        self.assertIsNotNone(self.client.auth_client.auth)

    def test_login_successful_fully_attenuated(self):
        self.client.login(
            email="dummy email",
            password="test correct password",
            packages=[{"name": "foo", "series": "16"}],
            channels=["edge"],
            save=False,
        )
        # Client configuration is filled, but it's not saved on disk.
        self.assertThat(
            self.client.auth_client._conf._get_config_path(), Not(FileExists())
        )
        self.assertIsNotNone(self.client.auth_client.auth)

    def test_login_successful_with_expiration(self):
        self.client.login(
            email="dummy email",
            password="test correct password",
            packages=[{"name": "foo", "series": "16"}],
            channels=["edge"],
            expires="2017-12-22",
        )
        self.assertIsNotNone(self.client.auth_client.auth)

    def test_login_with_exported_login(self):
        with pathlib.Path("test-exported-login").open("w") as config_fd:
            print(
                "[{}]".format(self.client.auth_client._conf._get_section_name()),
                file=config_fd,
            )
            print(
                "macaroon=MDAwZWxvY2F0aW9uIAowMDEwaWRlbnRpZmllciAKMDAxNGNpZCB0ZXN0IGNhdmVhdAowMDE5dmlkIHRlc3QgdmVyaWZpYWNpb24KMDAxN2NsIGxvY2FsaG9zdDozNTM1MQowMDBmc2lnbmF0dXJlIAo",
                file=config_fd,
            )
            print(
                "unbound_discharge=MDAwZWxvY2F0aW9uIAowMDEwaWRlbnRpZmllciAKMDAwZnNpZ25hdHVyZSAK",
                file=config_fd,
            )
            config_fd.flush()

        with pathlib.Path("test-exported-login").open() as config_fd:
            self.client.login(config_fd=config_fd)

        self.assertThat(
            self.client.auth_client._conf.get("macaroon"),
            Equals(
                "MDAwZWxvY2F0aW9uIAowMDEwaWRlbnRpZmllciAKMDAxNGNpZCB0ZXN0IGNhdmVhdAowMDE5dmlkIHRlc3QgdmVyaWZpYWNpb24KMDAxN2NsIGxvY2FsaG9zdDozNTM1MQowMDBmc2lnbmF0dXJlIAo"
            ),
        )
        self.assertThat(
            self.client.auth_client._conf.get("unbound_discharge"),
            Equals("MDAwZWxvY2F0aW9uIAowMDEwaWRlbnRpZmllciAKMDAwZnNpZ25hdHVyZSAK"),
        )
        self.assertThat(
            self.client.auth_client.auth,
            Equals(
                "Macaroon root=MDAwZWxvY2F0aW9uIAowMDEwaWRlbnRpZmllciAKMDAxNGNpZCB0ZXN0IGNhdmVhdAowMDE5dmlkIHRlc3QgdmVyaWZpYWNpb24KMDAxN2NsIGxvY2FsaG9zdDozNTM1MQowMDBmc2lnbmF0dXJlIAo, discharge=MDAwZWxvY2F0aW9uIAowMDEwaWRlbnRpZmllciAKMDAyZnNpZ25hdHVyZSDmRizXTOkAmfmy5hGCm7F0H4LBea16YbJYVhDkAJZ-Ago"
            ),
        )

    def test_failed_login_with_wrong_password(self):
        self.assertRaises(
            http_clients.errors.StoreAuthenticationError,
            self.client.login,
            email="dummy email",
            password="wrong password",
        )

    def test_failed_login_requires_one_time_password(self):
        self.assertRaises(
            http_clients.errors.StoreTwoFactorAuthenticationRequired,
            self.client.login,
            email="dummy email",
            password="test requires 2fa",
        )

    def test_failed_login_with_wrong_one_time_password(self):
        self.assertRaises(
            http_clients.errors.StoreAuthenticationError,
            self.client.login,
            email="dummy email",
            password="test correct password",
            otp="wrong one-time password",
        )

    def test_failed_login_with_unregistered_snap(self):
        raised = self.assertRaises(
            errors.GeneralStoreError,
            self.client.login,
            email="dummy email",
            password="test correct password",
            packages=[{"name": "unregistered-snap-name", "series": "16"}],
        )

        self.assertThat(str(raised), Contains("not found"))


class DownloadTestCase(StoreTestCase):

    # sha3-384 of tests/data/test-snap.snap
    EXPECTED_SHA3_384 = ""

    def test_download_nonexistent_snap_raises_exception(self):
        self.client.login(email="dummy", password="test correct password")

        raised = self.assertRaises(
            errors.SnapNotFoundError,
            self.client.download,
            "nonexistent-snap",
            risk="stable",
            download_path="dummy.snap",
            arch="test-arch",
        )

        self.expectThat(raised._snap_name, Equals("nonexistent-snap"))
        self.expectThat(raised._channel, Is(None))
        self.expectThat(raised._arch, Is(None))

    def test_download_snap(self):
        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)
        self.client.login(email="dummy", password="test correct password")
        download_path = os.path.join(self.path, "test-snap.snap")
        self.client.download("test-snap", risk="stable", download_path=download_path)
        self.assertThat(download_path, FileExists())

    def test_download_snap_missing_risk(self):
        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)
        self.client.login(email="dummy", password="test correct password")

        raised = self.assertRaises(
            errors.SnapNotFoundError,
            self.client.download,
            "test-snap",
            risk="beta",
            download_path="dummy.snap",
        )

        self.expectThat(raised._snap_name, Equals("test-snap"))
        self.expectThat(raised._channel, Equals("beta"))
        self.expectThat(raised._arch, Is(None))

    def test_download_from_brand_store_requires_store(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.SnapNotFoundError,
            self.client.download,
            "test-snap-brand-store",
            risk="stable",
            download_path="dummy.snap",
        )

        self.expectThat(raised._snap_name, Equals("test-snap-brand-store"))
        self.expectThat(raised._channel, Is(None))
        self.expectThat(raised._arch, Is(None))

    def test_download_from_branded_store(self):
        # Downloading from a branded-store requires setting the
        # 'SNAPCRAFT_UBUNTU_STORE' environment variable to the
        # correct store 'slug' (the branded store identifier).
        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)

        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_UBUNTU_STORE", "Test-Branded")
        )
        self.client.login(email="dummy", password="test correct password")

        download_path = os.path.join(self.path, "brand.snap")
        self.client.download(
            "test-snap-branded-store", risk="stable", download_path=download_path
        )
        self.assertThat(download_path, FileExists())

    def test_download_already_downloaded_snap(self):
        self.client.login(email="dummy", password="test correct password")
        download_path = os.path.join(self.path, "test-snap.snap")
        # download first time.
        self.client.download("test-snap", risk="stable", download_path=download_path)
        first_stat = os.stat(download_path)
        # download again.
        self.client.download("test-snap", risk="stable", download_path=download_path)
        second_stat = os.stat(download_path)
        # If these are equal it means a second download did not happen.
        self.assertThat(second_stat.st_ctime, Equals(first_stat.st_ctime))

    def test_download_on_sha_mismatch(self):
        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)
        self.client.login(email="dummy", password="test correct password")
        download_path = os.path.join(self.path, "test-snap.snap")
        # Write a wrong file in the download path.
        open(download_path, "w").close()
        first_stat = os.stat(download_path)
        self.client.download("test-snap", risk="stable", download_path=download_path)
        second_stat = os.stat(download_path)
        # If these are different it means that the download did happen.
        self.assertThat(second_stat, Not(Equals(first_stat)))

    def test_download_with_hash_mismatch_raises_exception(self):
        self.client.login(email="dummy", password="test correct password")
        download_path = os.path.join(self.path, "test-snap.snap")
        self.assertRaises(
            errors.SHAMismatchError,
            self.client.download,
            "test-snap-with-wrong-sha",
            risk="stable",
            download_path=download_path,
        )


class PushSnapBuildTestCase(StoreTestCase):
    def test_push_snap_build_refreshes_macaroon(self):
        self.client.login(email="dummy", password="test correct password")
        self.fake_store.needs_refresh = True
        self.client.push_snap_build("snap-id", "dummy")
        self.assertFalse(self.fake_store.needs_refresh)

    def test_push_snap_build_not_implemented(self):
        # If the "enable_snap_build" feature switch is off in the store, we
        # will get a descriptive error message.
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            http_clients.errors.StoreServerError,
            self.client.push_snap_build,
            "snap-id",
            "test-not-implemented",
        )
        self.assertThat(raised.error_code, Equals(501))

    def test_push_snap_build_invalid_data(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreSnapBuildError,
            self.client.push_snap_build,
            "snap-id",
            "test-invalid-data",
        )
        self.assertThat(
            str(raised),
            Equals("Could not assert build: The snap-build assertion is not " "valid."),
        )

    def test_push_snap_build_unexpected_data(self):
        # The endpoint in SCA would never return plain/text, however anything
        # might happen in the internet, so we are a little defensive.
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            http_clients.errors.StoreServerError,
            self.client.push_snap_build,
            "snap-id",
            "test-unexpected-data",
        )
        self.assertThat(raised.error_code, Equals(500))

    def test_push_snap_build_successfully(self):
        self.client.login(email="dummy", password="test correct password")
        # No exception will be raised if this is successful.
        self.client.push_snap_build("snap-id", "dummy")


class GetAccountInformationTestCase(StoreTestCase):
    def test_get_account_information_successfully(self):
        self.client.login(email="dummy", password="test correct password")
        self.assertThat(
            self.client.get_account_information(),
            Equals(
                {
                    "account_id": "abcd",
                    "account_keys": [],
                    "snaps": {
                        "16": {
                            "basic": {
                                "snap-id": "snap-id",
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                            "core": {
                                "snap-id": "good",
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                            "core-no-dev": {
                                "snap-id": "no-dev",
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                            "badrequest": {
                                "snap-id": "badrequest",
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                            "no-revoked": {
                                "snap-id": "no-revoked",
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                            "revoked": {
                                "snap-id": "revoked",
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                            "test-snap-with-dev": {
                                "price": None,
                                "private": False,
                                "since": "2016-12-12T01:01:01Z",
                                "snap-id": "test-snap-id-with-dev",
                                "status": "Approved",
                            },
                            "test-snap-with-no-validations": {
                                "price": None,
                                "private": False,
                                "since": "2016-12-12T01:01:01Z",
                                "snap-id": "test-snap-id-with-no-validations",
                                "status": "Approved",
                            },
                            "no-id": {
                                "snap-id": None,
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                        }
                    },
                }
            ),
        )

    def test_get_account_information_refreshes_macaroon(self):
        self.client.login(email="dummy", password="test correct password")
        self.fake_store.needs_refresh = True
        self.assertThat(
            self.client.get_account_information(),
            Equals(
                {
                    "account_id": "abcd",
                    "account_keys": [],
                    "snaps": {
                        "16": {
                            "basic": {
                                "snap-id": "snap-id",
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                            "core": {
                                "snap-id": "good",
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                            "core-no-dev": {
                                "snap-id": "no-dev",
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                            "badrequest": {
                                "snap-id": "badrequest",
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                            "no-revoked": {
                                "snap-id": "no-revoked",
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                            "revoked": {
                                "snap-id": "revoked",
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                            "test-snap-with-dev": {
                                "price": None,
                                "private": False,
                                "since": "2016-12-12T01:01:01Z",
                                "snap-id": "test-snap-id-with-dev",
                                "status": "Approved",
                            },
                            "test-snap-with-no-validations": {
                                "price": None,
                                "private": False,
                                "since": "2016-12-12T01:01:01Z",
                                "snap-id": "test-snap-id-with-no-validations",
                                "status": "Approved",
                            },
                            "no-id": {
                                "snap-id": None,
                                "status": "Approved",
                                "private": False,
                                "price": None,
                                "since": "2016-12-12T01:01:01Z",
                            },
                        }
                    },
                }
            ),
        )
        self.assertFalse(self.fake_store.needs_refresh)


class RegisterKeyTestCase(StoreTestCase):
    def test_register_key_successfully(self):
        self.client.login(email="dummy", password="test correct password")
        # No exception will be raised if this is successful.
        self.client.register_key(
            dedent(
                """\
            name: default
            public-key-sha3-384: abcd
            """
            )
        )

    def test_register_key_refreshes_macaroon(self):
        self.client.login(email="dummy", password="test correct password")
        self.fake_store.needs_refresh = True
        self.client.register_key(
            dedent(
                """\
            name: default
            public-key-sha3-384: abcd
            """
            )
        )
        self.assertFalse(self.fake_store.needs_refresh)

    def test_not_implemented(self):
        # If the enable_account_key feature switch is off in the store, we
        # will get a 501 Not Implemented response.
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            http_clients.errors.StoreServerError,
            self.client.register_key,
            "test-not-implemented",
        )
        self.assertThat(raised.error_code, Equals(501))

    def test_invalid_data(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreKeyRegistrationError,
            self.client.register_key,
            "test-invalid-data",
        )
        self.assertThat(
            str(raised),
            Equals(
                "Key registration failed: "
                "The account-key-request assertion is not valid."
            ),
        )


class RegisterTestCase(StoreTestCase):
    def test_register_name_successfully(self):
        self.client.login(email="dummy", password="test correct password")
        # No exception will be raised if this is successful
        self.client.register("test-good-snap-name")

    def test_register_name_successfully_to_store_id(self):
        self.client.login(email="dummy", password="test correct password")
        # No exception will be raised if this is successful
        self.client.register("test-good-snap-name", store_id="my-brand")

    def test_register_private_name_successfully(self):
        self.client.login(email="dummy", password="test correct password")
        # No exception will be raised if this is successful
        self.client.register("test-good-snap-name", is_private=True)

    def test_register_refreshes_macaroon(self):
        self.client.login(email="dummy", password="test correct password")
        self.fake_store.needs_refresh = True
        self.client.register("test-good-snap-name")
        self.assertFalse(self.fake_store.needs_refresh)

    def test_already_registered(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreRegistrationError,
            self.client.register,
            "test-snap-name-already-registered",
        )
        self.assertThat(
            str(raised),
            Equals(
                "The name 'test-snap-name-already-registered' is already "
                "taken.\n\n"
                "We can if needed rename snaps to ensure they match the "
                "expectations of most users. If you are the publisher most "
                "users expect for 'test-snap-name-already-registered' then "
                "claim the name at 'https://myapps.com/register-name/'"
            ),
        )

    def test_register_a_reserved_name(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreRegistrationError,
            self.client.register,
            "test-reserved-snap-name",
        )
        self.assertThat(
            str(raised),
            Equals(
                "The name 'test-reserved-snap-name' is reserved."
                "\n\n"
                "If you are the publisher most users expect for "
                "'test-reserved-snap-name' then please claim the "
                "name at 'https://myapps.com/register-name/'\n\n"
                "Otherwise, please register another name."
            ),
        )

    def test_register_already_owned_name(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreRegistrationError,
            self.client.register,
            "test-already-owned-snap-name",
        )
        self.assertThat(
            str(raised),
            Equals("You already own the name 'test-already-owned-snap-name'."),
        )

    def test_registering_too_fast_in_a_row(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreRegistrationError, self.client.register, "test-snapcraft-fast"
        )
        self.assertThat(
            str(raised),
            Equals(
                "You must wait 177 seconds before trying to register your " "next snap."
            ),
        )

    def test_registering_name_too_long(self):
        self.client.login(email="dummy", password="test correct password")
        name = "name-too-l{}ng".format("0" * 40)
        raised = self.assertRaises(
            errors.StoreRegistrationError, self.client.register, name
        )
        expected = (
            "The name '{}' is not valid: it should be no longer than 40"
            " characters.".format(name)
        )
        self.assertThat(str(raised), Equals(expected))

    def test_registering_name_invalid(self):
        self.client.login(email="dummy", password="test correct password")
        name = "test_invalid"
        raised = self.assertRaises(
            errors.StoreRegistrationError, self.client.register, name
        )
        expected = (
            "The name '{}' is not valid: it should only have"
            " ASCII lowercase letters, numbers, and hyphens,"
            " and must have at least one letter.".format(name)
        )
        self.assertThat(str(raised), Equals(expected))

    def test_unhandled_registration_error_path(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreRegistrationError,
            self.client.register,
            "snap-name-no-clear-error",
        )
        self.assertThat(str(raised), Equals("Registration failed."))


class ValidationSetsTestCase(StoreTestCase):
    def setUp(self):
        super().setUp()
        self.client.login(email="dummy", password="test correct password")

        self.validation_sets_build = {
            "name": "acme-cert-2020-10",
            "account-id": "AccountIDXXXOfTheRequestingUserX",
            "sequence": 3,
            "snaps": [
                {"name": "snap-name-1", "presence": "optional"},
                {"name": "snap-name-2"},
                {
                    "name": "snap-name-3",
                    "id": "XXSnapIDForXSnapName3XXXXXXXXXXX",
                    "presence": "required",
                    "revision": 2,
                },
                {"name": "snap-name-4", "revision": 123},
                {"name": "snap-name-5", "presence": "invalid"},
            ],
        }

    def test_post_valid_build_assertion(self):
        build_assertion = self.client.post_validation_sets_build_assertion(
            validation_sets=self.validation_sets_build
        )

        self.assertThat(build_assertion, IsInstance(validation_sets.BuildAssertion))

    def test_post_invalid_sequence_for_build_assertion(self):
        self.validation_sets_build["sequence"] = 0

        raised = self.assertRaises(
            errors.StoreValidationSetsError,
            self.client.post_validation_sets_build_assertion,
            validation_sets=self.validation_sets_build,
        )

        self.assertThat(
            str(raised),
            Equals(
                "Issues encountered with validation set: 0 is less than the minimum of 1 at /sequence"
            ),
        )

    def _fake_sign(self, build_assertion: validation_sets.BuildAssertion) -> bytes:
        # Fake sign.
        assertion_json = build_assertion.marshal()
        assertion_json[
            "sign-key-sha3-384"
        ] = "XSignXKeyXHashXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"
        return (json.dumps(assertion_json) + "\n\nSIGNED").encode()

    def test_post_valid_assertion(self):
        build_assertion = self.client.post_validation_sets_build_assertion(
            validation_sets=self.validation_sets_build
        )

        vs = self.client.post_validation_sets(
            signed_validation_sets=self._fake_sign(build_assertion)
        )

        self.assertThat(vs, IsInstance(validation_sets.ValidationSets))
        self.assertThat(vs.assertions, HasLength(1))
        self.assertThat(vs.assertions[0], Equals(build_assertion))

    def test_post_invalid_assertion(self):
        build_assertion = self.client.post_validation_sets_build_assertion(
            validation_sets=self.validation_sets_build
        )
        build_assertion.authority_id = "invalid"

        raised = self.assertRaises(
            errors.StoreValidationSetsError,
            self.client.post_validation_sets,
            signed_validation_sets=self._fake_sign(build_assertion),
        )

        self.assertThat(
            str(raised),
            Equals(
                "Issues encountered with validation set: account-id and authority-id must match the requesting user."
            ),
        )

    def test_get_validation_sets_empty(self):
        vs = self.client.get_validation_sets()

        self.assertThat(vs, IsInstance(validation_sets.ValidationSets))
        self.assertThat(vs.assertions, HasLength(0))

    def test_get_validation_sets(self):
        build_assertion = self.client.post_validation_sets_build_assertion(
            validation_sets=self.validation_sets_build
        )
        vs_1 = self.client.post_validation_sets(
            signed_validation_sets=self._fake_sign(build_assertion)
        )

        vs = self.client.get_validation_sets()

        self.assertThat(vs, IsInstance(validation_sets.ValidationSets))
        self.assertThat(vs.assertions, HasLength(1))
        self.expectThat(vs.assertions, Contains(vs_1.assertions[0]))

    def test_get_multiple_validation_sets(self):
        build_assertion = self.client.post_validation_sets_build_assertion(
            validation_sets=self.validation_sets_build
        )
        vs_1 = self.client.post_validation_sets(
            signed_validation_sets=self._fake_sign(build_assertion)
        )

        # Create a different build_assertion from the first one.
        build_assertion.name = "not-acme"
        vs_2 = self.client.post_validation_sets(
            signed_validation_sets=self._fake_sign(build_assertion)
        )

        vs = self.client.get_validation_sets()

        self.assertThat(vs, IsInstance(validation_sets.ValidationSets))
        self.expectThat(vs.assertions, Contains(vs_1.assertions[0]))
        self.expectThat(vs.assertions, Contains(vs_2.assertions[0]))

    def test_get_validation_sets_by_name(self):
        build_assertion = self.client.post_validation_sets_build_assertion(
            validation_sets=self.validation_sets_build
        )
        self.client.post_validation_sets(
            signed_validation_sets=self._fake_sign(build_assertion)
        )

        # Create a different build_assertion from the first one.
        build_assertion.name = "not-acme"
        self.client.post_validation_sets(
            signed_validation_sets=self._fake_sign(build_assertion)
        )
        vs = self.client.get_validation_sets(name="acme-cert-2020-10")

        self.assertThat(vs, IsInstance(validation_sets.ValidationSets))
        self.expectThat(vs.assertions, HasLength(1))
        self.expectThat(vs.assertions[0].name, Equals("acme-cert-2020-10"))

    def test_get_invalid_sequence(self):
        raised = self.assertRaises(
            errors.StoreValidationSetsError,
            self.client.get_validation_sets,
            sequence="invalid",
        )

        self.assertThat(
            str(raised),
            Equals(
                "Issues encountered with validation set: sequence must be one of "
                "['all', 'latest'] or a positive integer higher than 0."
            ),
        )


class ValidationsTestCase(StoreTestCase):
    def setUp(self):
        super().setUp()
        self.fake_logger = fixtures.FakeLogger(level=logging.DEBUG)
        self.useFixture(self.fake_logger)

    def test_get_success(self):
        self.client.login(email="dummy", password="test correct password")
        expected = [
            {
                "approved-snap-id": "snap-id-1",
                "approved-snap-revision": "3",
                "approved-snap-name": "snap-1",
                "authority-id": "dev-1",
                "series": "16",
                "sign-key-sha3-384": "1234567890",
                "snap-id": "snap-id-gating",
                "timestamp": "2016-09-19T21:07:27.756001Z",
                "type": "validation",
                "revoked": "false",
                "required": True,
            },
            {
                "approved-snap-id": "snap-id-2",
                "approved-snap-revision": "5",
                "approved-snap-name": "snap-2",
                "authority-id": "dev-1",
                "series": "16",
                "sign-key-sha3-384": "1234567890",
                "snap-id": "snap-id-gating",
                "timestamp": "2016-09-19T21:07:27.756001Z",
                "type": "validation",
                "revoked": "false",
                "required": False,
            },
            {
                "approved-snap-id": "snap-id-3",
                "approved-snap-revision": "-",
                "approved-snap-name": "snap-3",
                "authority-id": "dev-1",
                "series": "16",
                "sign-key-sha3-384": "1234567890",
                "snap-id": "snap-id-gating",
                "timestamp": "2016-09-19T21:07:27.756001Z",
                "type": "validation",
                "revoked": "false",
                "required": True,
            },
        ]
        result = self.client.get_assertion("good", "validations")
        self.assertThat(result, Equals(expected))

    def test_get_bad_response(self):
        self.client.login(email="dummy", password="test correct password")

        err = self.assertRaises(
            errors.StoreValidationError, self.client.get_assertion, "bad", "validations"
        )

        expected = "Received error 200: 'Invalid response from the server'"
        self.assertThat(str(err), Equals(expected))
        self.assertIn("Invalid response from the server", self.fake_logger.output)

    def test_get_error_response(self):
        self.client.login(email="dummy", password="test correct password")

        err = self.assertRaises(
            http_clients.errors.StoreNetworkError,
            self.client.get_assertion,
            "err",
            "validations",
        )

        expected = "maximum retries exceeded"
        self.assertThat(str(err), Contains(expected))

    def test_push_success(self):
        self.client.login(email="dummy", password="test correct password")
        assertion = json.dumps({"foo": "bar"}).encode("utf-8")

        result = self.client.push_assertion("good", assertion, "validations")

        expected = {"assertion": '{"foo": "bar"}'}
        self.assertThat(result, Equals(expected))

    def test_push_bad_response(self):
        self.client.login(email="dummy", password="test correct password")
        assertion = json.dumps({"foo": "bar"}).encode("utf-8")

        err = self.assertRaises(
            errors.StoreValidationError,
            self.client.push_assertion,
            "bad",
            assertion,
            "validations",
        )

        expected = "Received error 200: 'Invalid response from the server'"
        self.assertThat(str(err), Equals(expected))
        self.assertIn("Invalid response from the server", self.fake_logger.output)

    def test_push_error_response(self):
        self.client.login(email="dummy", password="test correct password")
        assertion = json.dumps({"foo": "bar"}).encode("utf-8")

        err = self.assertRaises(
            http_clients.errors.StoreServerError,
            self.client.push_assertion,
            "err",
            assertion,
            "validations",
        )
        self.assertThat(err.error_code, Equals(501))


class UploadTestCase(StoreTestCase):
    def setUp(self):
        super().setUp()
        self.snap_path = os.path.join(
            os.path.dirname(tests.__file__), "data", "test-snap.snap"
        )
        # These should eventually converge to the same module
        pbars = (
            "snapcraft.storeapi._upload.ProgressBar",
            "snapcraft.storeapi._status_tracker.ProgressBar",
        )
        for pbar in pbars:
            patcher = mock.patch(pbar, new=unit.SilentProgressBar)
            patcher.start()
            self.addCleanup(patcher.stop)

    def test_upload_snap(self):
        self.client.login(email="dummy", password="test correct password")
        self.client.register("test-snap")
        tracker = self.client.upload("test-snap", self.snap_path)
        self.assertTrue(isinstance(tracker, storeapi._status_tracker.StatusTracker))
        result = tracker.track()
        expected_result = {
            "code": "ready_to_release",
            "revision": "1",
            "url": "/dev/click-apps/5349/rev/1",
            "can_release": True,
            "processed": True,
        }
        self.assertThat(result, Equals(expected_result))

        # This should not raise
        tracker.raise_for_code()

    def test_upload_refreshes_macaroon(self):
        self.client.login(email="dummy", password="test correct password")
        self.client.register("test-snap")
        self.fake_store.needs_refresh = True
        tracker = self.client.upload("test-snap", self.snap_path)
        result = tracker.track()
        expected_result = {
            "code": "ready_to_release",
            "revision": "1",
            "url": "/dev/click-apps/5349/rev/1",
            "can_release": True,
            "processed": True,
        }
        self.assertThat(result, Equals(expected_result))

        # This should not raise
        tracker.raise_for_code()

        self.assertFalse(self.fake_store.needs_refresh)

    def test_upload_snap_fails_due_to_upload_fail(self):
        # Tells the fake updown server to return a 5xx response
        self.useFixture(fixtures.EnvironmentVariable("UPDOWN_BROKEN", "1"))

        self.client.login(email="dummy", password="test correct password")

        raised = self.assertRaises(
            http_clients.errors.StoreServerError,
            self.client.upload,
            "test-snap",
            self.snap_path,
        )
        self.assertThat(raised.error_code, Equals(500))

    def test_upload_snap_requires_review(self):
        self.client.login(email="dummy", password="test correct password")
        self.client.register("test-review-snap")
        tracker = self.client.upload("test-review-snap", self.snap_path)
        self.assertTrue(isinstance(tracker, storeapi._status_tracker.StatusTracker))
        result = tracker.track()
        expected_result = {
            "code": "need_manual_review",
            "revision": "1",
            "url": "/dev/click-apps/5349/rev/1",
            "can_release": False,
            "processed": True,
        }
        self.assertThat(result, Equals(expected_result))

        self.assertRaises(errors.StoreReviewError, tracker.raise_for_code)

    def test_upload_duplicate_snap(self):
        self.client.login(email="dummy", password="test correct password")
        self.client.register("test-duplicate-snap")
        tracker = self.client.upload("test-duplicate-snap", self.snap_path)
        self.assertTrue(isinstance(tracker, storeapi._status_tracker.StatusTracker))
        result = tracker.track()
        expected_result = {
            "code": "processing_error",
            "revision": "1",
            "url": "/dev/click-apps/5349/rev/1",
            "can_release": False,
            "processed": True,
            "errors": [{"message": "Duplicate snap already uploaded"}],
        }
        self.assertThat(result, Equals(expected_result))

        raised = self.assertRaises(errors.StoreReviewError, tracker.raise_for_code)

        self.assertThat(
            str(raised),
            Equals(
                "The store was unable to accept this snap.\n"
                "  - Duplicate snap already uploaded"
            ),
        )

    def test_braces_in_error_messages_are_literals(self):
        self.client.login(email="dummy", password="test correct password")
        self.client.register("test-scan-error-with-braces")
        tracker = self.client.upload("test-scan-error-with-braces", self.snap_path)
        self.assertTrue(isinstance(tracker, storeapi._status_tracker.StatusTracker))
        result = tracker.track()
        expected_result = {
            "code": "processing_error",
            "revision": "1",
            "url": "/dev/click-apps/5349/rev/1",
            "can_release": False,
            "processed": True,
            "errors": [{"message": "Error message with {braces}"}],
        }
        self.assertThat(result, Equals(expected_result))

        raised = self.assertRaises(errors.StoreReviewError, tracker.raise_for_code)

        self.assertThat(
            str(raised),
            Equals(
                "The store was unable to accept this snap.\n"
                "  - Error message with {braces}"
            ),
        )

    def test_upload_unregistered_snap(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreUploadError,
            self.client.upload,
            "test-snap-unregistered",
            self.snap_path,
        )
        self.assertThat(
            str(raised),
            Equals("This snap is not registered. Register the snap and try again."),
        )

    def test_upload_forbidden_snap(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreUploadError,
            self.client.upload,
            "test-snap-forbidden",
            self.snap_path,
        )
        self.assertThat(
            str(raised),
            Equals(
                "You are not the publisher or allowed to upload revisions for "
                "this snap. Ensure you are logged in with the proper account "
                "and try again."
            ),
        )


class ReleaseTest(StoreTestCase):
    def test_release_snap(self):
        self.client.login(email="dummy", password="test correct password")
        channel_map = self.client.release("test-snap", "19", ["beta"])
        expected_channel_map = {
            "opened_channels": ["beta"],
            "channel_map": [
                {"channel": "stable", "info": "none"},
                {"channel": "candidate", "info": "none"},
                {"revision": 19, "channel": "beta", "version": "0", "info": "specific"},
                {"channel": "edge", "info": "tracking"},
            ],
        }
        self.assertThat(channel_map, Equals(expected_channel_map))

    def test_progressive_release_snap(self):
        self.client.login(email="dummy", password="test correct password")
        channel_map = self.client.release(
            "test-snap", "19", ["beta"], progressive_percentage=10
        )
        expected_channel_map = {
            "opened_channels": ["beta"],
            "channel_map": [
                {"channel": "stable", "info": "none"},
                {"channel": "candidate", "info": "none"},
                {"revision": 19, "channel": "beta", "version": "0", "info": "specific"},
                {"channel": "edge", "info": "tracking"},
            ],
        }
        # The channel map would be the same as if no progressive release was
        # done.
        self.assertThat(channel_map, Equals(expected_channel_map))

    def test_release_refreshes_macaroon(self):
        self.client.login(email="dummy", password="test correct password")
        self.fake_store.needs_refresh = True
        channel_map = self.client.release("test-snap", "19", ["beta"])
        expected_channel_map = {
            "opened_channels": ["beta"],
            "channel_map": [
                {"channel": "stable", "info": "none"},
                {"channel": "candidate", "info": "none"},
                {"revision": 19, "channel": "beta", "version": "0", "info": "specific"},
                {"channel": "edge", "info": "tracking"},
            ],
        }
        self.assertThat(channel_map, Equals(expected_channel_map))
        self.assertFalse(self.fake_store.needs_refresh)

    def test_release_snap_to_invalid_channel(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreReleaseError, self.client.release, "test-snap", "19", ["alpha"]
        )

        self.assertThat(str(raised), Equals("Not a valid channel: alpha"))

    def test_release_snap_to_bad_channel(self):
        self.client.login(email="dummy", password="test correct password")
        self.assertRaises(
            http_clients.errors.StoreServerError,
            self.client.release,
            "test-snap",
            "19",
            ["bad-channel"],
        )

    def test_release_unregistered_snap(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreReleaseError,
            self.client.release,
            "test-snap-unregistered",
            "19",
            ["alpha"],
        )

        self.assertThat(
            str(raised),
            Equals(
                "Sorry, try `snapcraft register test-snap-unregistered` "
                "before trying to release or choose an existing "
                "revision."
            ),
        )

    def test_release_with_invalid_revision(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreReleaseError,
            self.client.release,
            "test-snap-invalid-data",
            "notanumber",
            ["beta"],
        )

        self.assertThat(
            str(raised),
            Equals("invalid-field: The 'revision' field must be an integer"),
        )

    def test_release_to_curly_braced_channel(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreReleaseError,
            self.client.release,
            "test-snap",
            "19",
            ["edge/{curly}"],
        )

        self.assertThat(
            str(raised),
            Equals(
                "invalid-field: Invalid branch name: {curly}. Enter a value consisting of "
                "letters, numbers or hyphens. Hyphens cannot occur at the start or end of the "
                "chosen value."
            ),
        )


class CloseChannelsTestCase(StoreTestCase):
    def setUp(self):
        super().setUp()
        self.fake_logger = fixtures.FakeLogger(level=logging.DEBUG)
        self.useFixture(self.fake_logger)

    def test_close_refreshes_macaroon(self):
        self.client.login(email="dummy", password="test correct password")
        self.fake_store.needs_refresh = True
        self.client.close_channels("snap-id", ["dummy"])
        self.assertFalse(self.fake_store.needs_refresh)

    def test_close_invalid_data(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreChannelClosingError,
            self.client.close_channels,
            "snap-id",
            ["invalid"],
        )
        self.assertThat(
            str(raised),
            Equals(
                "Could not close channel: The 'channels' field content " "is not valid."
            ),
        )

    def test_close_unexpected_data(self):
        # The endpoint in SCA would never return plain/text, however anything
        # might happen in the internet, so we are a little defensive.
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            http_clients.errors.StoreServerError,
            self.client.close_channels,
            "snap-id",
            ["unexpected"],
        )
        self.assertThat(raised.error_code, Equals(500))

    def test_close_broken_store_plain(self):
        # If the contract is broken by the Store, users will be have additional
        # debug information available.
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreChannelClosingError,
            self.client.close_channels,
            "snap-id",
            ["broken-plain"],
        )
        self.assertThat(str(raised), Equals("Could not close channel: 200 OK"))

        expected_lines = [
            "Invalid response from the server on channel closing:",
            "200 OK",
            "b'plain data'",
        ]

        actual_lines = []
        for line in self.fake_logger.output.splitlines():
            line = line.strip()
            if line in expected_lines:
                actual_lines.append(line)

        self.assertThat(actual_lines, Equals(expected_lines))

    def test_close_broken_store_json(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.StoreChannelClosingError,
            self.client.close_channels,
            "snap-id",
            ["broken-json"],
        )
        self.assertThat(str(raised), Equals("Could not close channel: 200 OK"))

        expected_lines = [
            "Invalid response from the server on channel closing:",
            "200 OK",
            'b\'{"closed_channels": ["broken-json"]}\'',
        ]

        actual_lines = []
        for line in self.fake_logger.output.splitlines():
            line = line.strip()
            if line in expected_lines:
                actual_lines.append(line)

        self.assertThat(actual_lines, Equals(expected_lines))

    def test_close_successfully(self):
        # Successfully closing a channels returns 'closed_channels'
        # and 'channel_map_tree' from the Store.
        self.client.login(email="dummy", password="test correct password")
        closed_channels, channel_map_tree = self.client.close_channels(
            "snap-id", ["beta"]
        )
        self.assertThat(closed_channels, Equals(["beta"]))
        self.assertThat(
            channel_map_tree,
            Equals(
                {
                    "latest": {
                        "16": {
                            "amd64": [
                                {"channel": "stable", "info": "none"},
                                {"channel": "candidate", "info": "none"},
                                {
                                    "channel": "beta",
                                    "info": "specific",
                                    "revision": 42,
                                    "version": "1.1",
                                },
                                {"channel": "edge", "info": "tracking"},
                            ]
                        }
                    }
                }
            ),
        )


class GetSnapStatusTestCase(StoreTestCase):
    def setUp(self):
        super().setUp()
        self.expected = {
            "channel_map_tree": {
                "latest": {
                    "16": {
                        "i386": [
                            {"channel": "stable", "info": "none"},
                            {"channel": "candidate", "info": "none"},
                            {
                                "channel": "beta",
                                "info": "specific",
                                "revision": 6,
                                "version": "1.1-amd64",
                            },
                            {
                                "channel": "edge",
                                "info": "specific",
                                "revision": 3,
                                "version": "1.0-i386",
                            },
                            {
                                "channel": "edge/test",
                                "info": "branch",
                                "revision": 9,
                                "version": "1.1-i386",
                                "expires_at": "2019-05-30T01:17:06.465504",
                            },
                        ],
                        "all": [
                            {"channel": "stable", "info": "none"},
                            {"channel": "candidate", "info": "none"},
                            {
                                "channel": "beta",
                                "info": "specific",
                                "revision": 6,
                                "version": "1.1-amd64",
                            },
                            {
                                "channel": "edge",
                                "info": "specific",
                                "revision": 3,
                                "version": "1.0-i386",
                            },
                            {
                                "channel": "edge/test",
                                "info": "branch",
                                "revision": 9,
                                "version": "1.1-i386",
                                "expires_at": "2019-05-30T01:17:06.465504",
                            },
                        ],
                        "amd64": [
                            {
                                "channel": "stable",
                                "info": "specific",
                                "revision": 2,
                                "version": "1.0-amd64",
                            },
                            {"channel": "candidate", "info": "none"},
                            {
                                "channel": "beta",
                                "info": "specific",
                                "revision": 4,
                                "version": "1.1-amd64",
                            },
                            {"channel": "edge", "info": "tracking"},
                            {
                                "channel": "edge/test",
                                "info": "branch",
                                "revision": 10,
                                "version": "1.1-amd64",
                                "expires_at": "2019-05-30T01:17:06.465504",
                            },
                        ],
                    }
                }
            }
        }

    def test_get_snap_status_successfully(self):
        self.client.login(email="dummy", password="test correct password")
        self.assertThat(self.client.get_snap_status("basic"), Equals(self.expected))

    def test_get_snap_status_filter_by_arch(self):
        self.client.login(email="dummy", password="test correct password")
        exp_arch = self.expected["channel_map_tree"]["latest"]["16"]["amd64"]
        self.assertThat(
            self.client.get_snap_status("basic", arch="amd64"),
            Equals({"channel_map_tree": {"latest": {"16": {"amd64": exp_arch}}}}),
        )

    def test_get_snap_status_filter_by_unknown_arch(self):
        self.client.login(email="dummy", password="test correct password")

        raised = self.assertRaises(
            storeapi.errors.SnapNotFoundError,
            self.client.get_snap_status,
            "basic",
            arch="some-arch",
        )

        self.expectThat(raised._snap_name, Equals("basic"))
        self.expectThat(raised._channel, Is(None))
        self.expectThat(raised._arch, Is("some-arch"))

    def test_get_snap_status_no_id(self):
        self.client.login(email="dummy", password="test correct password")
        e = self.assertRaises(
            storeapi.errors.NoSnapIdError, self.client.get_snap_status, "no-id"
        )
        self.assertThat(e.snap_name, Equals("no-id"))

    def test_get_snap_status_refreshes_macaroon(self):
        self.client.login(email="dummy", password="test correct password")
        self.fake_store.needs_refresh = True
        self.assertThat(self.client.get_snap_status("basic"), Equals(self.expected))
        self.assertFalse(self.fake_store.needs_refresh)

    @mock.patch.object(storeapi.StoreClient, "get_account_information")
    @mock.patch.object(storeapi._dashboard_api.DashboardAPI, "get")
    def test_get_snap_status_server_error(self, mock_sca_get, mock_account_info):
        mock_account_info.return_value = {
            "snaps": {"16": {"basic": {"snap-id": "my_snap_id"}}}
        }

        mock_sca_get.return_value = mock.Mock(
            ok=False, status_code=500, reason="Server error", json=lambda: {}
        )

        self.client.login(email="dummy", password="test correct password")
        e = self.assertRaises(
            storeapi.errors.StoreSnapStatusError, self.client.get_snap_status, "basic"
        )
        self.assertThat(
            str(e),
            Equals(
                "Error fetching status of snap id 'my_snap_id' for 'any arch' "
                "in '16' series: 500 Server error."
            ),
        )


class SnapChannelMapTest(StoreTestCase):
    def test_get_snap_channel_map(self):
        self.client.login(email="dummy", password="test correct password")
        self.assertThat(
            self.client.get_snap_channel_map(snap_name="basic"),
            IsInstance(channel_map.ChannelMap),
        )


class SnapReleasesTest(StoreTestCase):
    def test_get_snap_releases(self):
        self.client.login(email="dummy", password="test correct password")
        self.assertThat(
            self.client.get_snap_releases(snap_name="basic"),
            IsInstance(releases.Releases),
        )


class WhoAmITest(StoreTestCase):
    def test_whoami(self):
        self.client.login(email="dummy", password="test correct password")
        self.assertThat(
            self.client.whoami(), IsInstance(whoami.WhoAmI),
        )


class SignDeveloperAgreementTestCase(StoreTestCase):
    def test_sign_dev_agreement_success(self):
        self.client.login(email="dummy", password="test correct password")
        response = {
            "content": {
                "latest_tos_accepted": True,
                "tos_url": "http://fake-url.com",
                "latest_tos_date": "2000-01-01",
                "accepted_tos_date": "2010-10-10",
            }
        }
        self.assertThat(
            response,
            Equals(self.client.sign_developer_agreement(latest_tos_accepted=True)),
        )

    def test_sign_dev_agreement_exception(self):
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            errors.DeveloperAgreementSignError,
            self.client.sign_developer_agreement,
            False,
        )
        self.assertIn(
            "There was an error while signing developer agreement.\n"
            "Reason: 'Bad Request'\n",
            str(raised),
        )

    def test_sign_dev_agreement_exception_store_down(self):
        self.useFixture(fixtures.EnvironmentVariable("STORE_DOWN", "1"))
        self.client.login(email="dummy", password="test correct password")
        raised = self.assertRaises(
            http_clients.errors.StoreServerError,
            self.client.sign_developer_agreement,
            latest_tos_accepted=True,
        )
        self.assertThat(raised.error_code, Equals(500))


class UploadMetadataTestCase(StoreTestCase):
    def setUp(self):
        super().setUp()
        self.fake_logger = fixtures.FakeLogger(level=logging.DEBUG)
        self.useFixture(self.fake_logger)

    def _setup_snap(self):
        """Login, register and upload a snap.

        These are all the previous steps needed to upload metadata.
        """
        self.client.login(email="dummy", password="test correct password")
        self.client.register("basic")
        path = os.path.join(os.path.dirname(tests.__file__), "data", "test-snap.snap")
        tracker = self.client.upload("basic", path)
        tracker.track()

    def test_refreshes_macaroon(self):
        self._setup_snap()
        self.fake_store.needs_refresh = True
        metadata = {"field_ok": "foo"}
        self.client.upload_metadata("basic", metadata, False)
        self.assertFalse(self.fake_store.needs_refresh)

    def test_invalid_data(self):
        self._setup_snap()
        metadata = {"invalid": "foo"}
        raised = self.assertRaises(
            errors.StoreMetadataError,
            self.client.upload_metadata,
            "basic",
            metadata,
            False,
        )
        self.assertThat(str(raised), Equals("Received 400: 'Invalid field: invalid'"))

    def test_all_ok(self):
        self._setup_snap()
        metadata = {"field_ok": "foo"}
        result = self.client.upload_metadata("basic", metadata, False)
        self.assertIsNone(result)

    def test_conflicting_simple_normal(self):
        self._setup_snap()
        metadata = {"test-conflict": "value"}
        raised = self.assertRaises(
            errors.StoreMetadataError,
            self.client.upload_metadata,
            "basic",
            metadata,
            False,
        )
        should = """
            Metadata not uploaded!
            Conflict in 'test-conflict' field:
                In snapcraft.yaml: 'value'
                In the Store:      'value-changed'
            You can repeat the upload-metadata command with --force to force the local values into the Store
        """  # NOQA
        self.assertThat(str(raised), Equals(dedent(should).strip()))

    def test_conflicting_multiple_normal(self):
        self._setup_snap()
        metadata = {"test-conflict-1": "value-1", "test-conflict-2": "value-2"}
        raised = self.assertRaises(
            errors.StoreMetadataError,
            self.client.upload_metadata,
            "basic",
            metadata,
            False,
        )
        should = """
            Metadata not uploaded!
            Conflict in 'test-conflict-1' field:
                In snapcraft.yaml: 'value-1'
                In the Store:      'value-1-changed'
            Conflict in 'test-conflict-2' field:
                In snapcraft.yaml: 'value-2'
                In the Store:      'value-2-changed'
            You can repeat the upload-metadata command with --force to force the local values into the Store
        """  # NOQA
        self.assertThat(str(raised), Equals(dedent(should).strip()))

    def test_conflicting_force(self):
        self._setup_snap()
        metadata = {"test-conflict": "value"}
        # force the update, even on conflicts!
        result = self.client.upload_metadata("basic", metadata, True)
        self.assertIsNone(result)

    def test_braces_in_error_messages_are_literals(self):
        self._setup_snap()
        metadata = {"test-conflict-with-braces": "value"}
        raised = self.assertRaises(
            errors.StoreMetadataError,
            self.client.upload_metadata,
            "basic",
            metadata,
            False,
        )
        should = """
            Metadata not uploaded!
            Conflict in 'test-conflict-with-braces' field:
                In snapcraft.yaml: 'value'
                In the Store:      'value with {braces}'
            You can repeat the upload-metadata command with --force to force the local values into the Store
        """  # NOQA
        self.assertThat(str(raised), Equals(dedent(should).strip()))


class UploadBinaryMetadataTestCase(StoreTestCase):
    def setUp(self):
        super().setUp()
        self.fake_logger = fixtures.FakeLogger(level=logging.DEBUG)
        self.useFixture(self.fake_logger)

    def _setup_snap(self):
        """Login, register and upload a snap.

        These are all the previous steps needed to upload binary metadata.
        """
        self.client.login(email="dummy", password="test correct password")
        self.client.register("basic")
        path = os.path.join(os.path.dirname(tests.__file__), "data", "test-snap.snap")
        tracker = self.client.upload("basic", path)
        tracker.track()

    def test_refreshes_macaroon(self):
        self._setup_snap()
        self.fake_store.needs_refresh = True
        with tempfile.NamedTemporaryFile(suffix="ok") as f:
            metadata = {"icon": f}
            self.client.upload_binary_metadata("basic", metadata, False)
        self.assertFalse(self.fake_store.needs_refresh)

    def test_invalid_data(self):
        self._setup_snap()
        with tempfile.NamedTemporaryFile(suffix="invalid") as f:
            metadata = {"icon": f}
            raised = self.assertRaises(
                errors.StoreMetadataError,
                self.client.upload_binary_metadata,
                "basic",
                metadata,
                False,
            )
        self.assertThat(str(raised), Equals("Received 400: 'Invalid field: icon'"))

    def test_all_ok(self):
        self._setup_snap()
        with tempfile.NamedTemporaryFile(suffix="ok") as f:
            metadata = {"icon": f}
            result = self.client.upload_binary_metadata("basic", metadata, False)
        self.assertIsNone(result)

    def test_conflicting_simple_normal(self):
        self._setup_snap()
        with tempfile.NamedTemporaryFile(suffix="conflict") as f:
            filename = os.path.basename(f.name)
            metadata = {"icon": f}
            raised = self.assertRaises(
                errors.StoreMetadataError,
                self.client.upload_binary_metadata,
                "basic",
                metadata,
                False,
            )
        should = """
            Metadata not uploaded!
            Conflict in 'icon' field:
                In snapcraft.yaml: '{}'
                In the Store:      'original-icon'
            You can repeat the upload-metadata command with --force to force the local values into the Store
        """.format(
            filename
        )  # NOQA
        self.assertThat(str(raised), Equals(dedent(should).strip()))

    def test_conflicting_force(self):
        self._setup_snap()
        with tempfile.NamedTemporaryFile(suffix="conflict") as f:
            metadata = {"icon": f}
            # force the update, even on conflicts!
            result = self.client.upload_binary_metadata("basic", metadata, True)
        self.assertIsNone(result)

    def test_braces_in_error_messages_are_literals(self):
        self._setup_snap()
        with tempfile.NamedTemporaryFile(suffix="conflict-with-braces") as f:
            filename = os.path.basename(f.name)
            metadata = {"icon": f}
            raised = self.assertRaises(
                errors.StoreMetadataError,
                self.client.upload_binary_metadata,
                "basic",
                metadata,
                False,
            )
        should = """
            Metadata not uploaded!
            Conflict in 'icon' field:
                In snapcraft.yaml: '{}'
                In the Store:      'original icon with {{braces}}'
            You can repeat the upload-metadata command with --force to force the local values into the Store
        """.format(
            filename
        )  # NOQA
        self.assertThat(str(raised), Equals(dedent(should).strip()))


class SnapNotFoundTestCase(StoreTestCase):
    def _setup_snap(self):
        """Login, register and upload a snap.

        These are all the previous steps needed to upload binary metadata.
        """
        self.client.login(email="dummy", password="test correct password")
        self.client.register("basic")
        path = os.path.join(os.path.dirname(tests.__file__), "data", "test-snap.snap")
        tracker = self.client.upload("basic", path)
        tracker.track()

    def assert_raises(self, method):
        self._setup_snap()

        metadata = {"field_ok": "dummy"}

        raised = self.assertRaises(
            errors.SnapNotFoundError,
            getattr(self.client, method),
            "test-nonexistent-snap",
            metadata,
            False,
        )

        self.expectThat(raised._snap_name, Equals("test-nonexistent-snap"))
        self.expectThat(raised._channel, Is(None))
        self.expectThat(raised._arch, Is(None))

    def test_snap_not_found_upload_metadata(self):
        self.assert_raises("upload_metadata")

    def test_snap_not_found_upload_binary_metadata(self):
        self.assert_raises("upload_binary_metadata")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import click

from snapcraft.internal import errors
from snapcraft.project._sanity_checks import conduct_project_sanity_check

from ._command import SnapcraftProjectCommand, run_legacy_snapcraft
from ._options import get_project


@click.group()
@click.pass_context
def legacycli(ctx, **kwargs):
    pass


@legacycli.command(
    cls=SnapcraftProjectCommand,
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
def cleanbuild():
    """Create a snap using a clean environment managed by a build provider.

    \b
    Examples:
        snapcraft cleanbuild

    The cleanbuild command requires a properly setup lxd environment that
    can connect to external networks. Refer to the "Ubuntu Desktop and
    Ubuntu Server" section on
    https://linuxcontainers.org/lxd/getting-started-cli
    to get started.

    If using a remote, a prior setup is required which is described on:
    https://linuxcontainers.org/lxd/getting-started-cli/#multiple-host

    This command is no longer available when using the base keyword."""
    # We can only end up here with an invalid yaml
    project = get_project()
    conduct_project_sanity_check(project)

    # if we got this far, the environment must be broken
    raise errors.SnapcraftEnvironmentError(
        "The cleanbuild command is no longer supported when using the base keyword."
    )


@legacycli.command(
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True)
)
def define():
    """Shows the definition for the cloud part.

    \b
    Examples:
        snapcraft define my-part1

    This command is no longer available when using the base keyword.
    """
    run_legacy_snapcraft()


@legacycli.command(
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True)
)
def update():
    """Updates the parts listing from the cloud.

    This command is no longer available when using the base keyword.
    """
    run_legacy_snapcraft()


@legacycli.command(
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True)
)
def search():
    """Searches the remote parts cache for matching parts.

    \b
    Examples:
        snapcraft search desktop

    This command is no longer available when using the base keyword.
    """
    run_legacy_snapcraft()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

import pytest

from snapcraft.internal.meta.snap import Snap
from snapcraft.project import Project


@pytest.fixture
def project_core18(monkeypatch, tmp_work_path):
    """Return project for core18"""
    monkeypatch.setattr(Project, "parallel_build_count", 2)

    snapcraft_project = Project()
    snapcraft_project._snap_meta = Snap(
        name="test-snap", base="core18", confinement="strict"
    )
    return snapcraft_project


@pytest.fixture(params=["core", "core18"])
def project(monkeypatch, tmp_work_path, request):
    """Return project variants for core and core18"""
    monkeypatch.setattr(Project, "parallel_build_count", 2)

    snapcraft_project = Project()
    snapcraft_project._snap_meta = Snap(
        name="test-snap", base=request.param, confinement="strict"
    )
    return snapcraft_project


@pytest.fixture
def mock_common_run_output():
    """A no-op common.run_output mock."""
    patcher = mock.patch("snapcraft.internal.common.run_output")
    yield patcher.start()
    patcher.stop()


@pytest.fixture
def mock_run():
    """A no-op run mock."""
    patcher = mock.patch("snapcraft.plugins.v1.PluginV1.run")
    yield patcher.start()
    patcher.stop()


@pytest.fixture
def mock_run_output():
    """A no-op run_output mock."""
    patcher = mock.patch("snapcraft.plugins.v1.PluginV1.run_output")
    yield patcher.start()
    patcher.stop()


@pytest.fixture
def mock_tar():
    """A no-op tar source mock."""
    patcher = mock.patch("snapcraft.internal.sources.Tar")
    yield patcher.start()
    patcher.stop()


@pytest.fixture
def mock_zip():
    """A no-op zip source mock."""
    patcher = mock.patch("snapcraft.internal.sources.Zip")
    yield patcher.start()
    patcher.stop()
